package de.uni_passau.spl.featurebite;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;

import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.ClassNode;

import de.uni_passau.spl.featurebite.adapter.composing.ComposingClassVisitor;
import de.uni_passau.spl.featurebite.adapter.rename.ClassRenameAdapter;
import de.uni_passau.spl.featurebite.adapter.rename.RenameUtil;
import de.uni_passau.spl.featurebite.spl.SPLStructure;
import de.uni_passau.spl.featurebite.stubs.StubUtil;
import de.uni_passau.spl.featurebite.util.ClassNodeTransformer;
import de.uni_passau.spl.featurebite.util.FilePathUtils;
import de.uni_passau.spl.featurebite.util.Timer;
import de.uni_passau.spl.featurebite.verification.BCELVerifier;
import de.uni_passau.spl.featurebite.verification.JarFileClassLoader;

/**
 * The {@link Composer} handles the composing algorithms for the underlying
 * {@link SPLStructure}. The mechanism is designed to behave like super
 * imposition via fuji - just for bytecode files.
 * 
 * @author Claus Hunsen
 */
public class Composer {

	/**
	 * The structure of the given product line.
	 */
	protected final SPLStructure spl;

	/**
	 * A list of class files on which renaming was made.
	 */
	protected Map<File, ClassWriter> renamedFiles = new HashMap<File, ClassWriter>();

	/**
	 * A list of finished class files within the composition folder.
	 */
	protected Map<File, ClassNode> finishedCompFiles = new HashMap<File, ClassNode>();

	// TODO doc
	private final static Logger logger = Logger.getLogger(Composer.class);

	// TODO doku
	private final String classpath;

	protected boolean stubsPresent;

	//

	/**
	 * Constructs a new composing unit that initializes, starts and delegates
	 * the composing mechanisms for the given {@link SPLStructure}.
	 * 
	 * @param spl
	 *            the {@link SPLStructure} that will be composed
	 * @param classpath
	 * 
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public Composer(SPLStructure spl, String classpath)
			throws FileNotFoundException, IOException {
		this.spl = spl;
		this.classpath = classpath;
		this.stubsPresent = false;

		this.spl.writeCompositionFiles();
	}

	/**
	 * Starts the composing process. Renames inner classes for structure
	 * purposes, composes the features and verifies the result and cleans up the
	 * working place.
	 * 
	 * @throws IOException
	 * @throws FileNotFoundException
	 * @throws IllegalStateException
	 * 
	 * @see #renameClasses()
	 * @see #composeFeatures()
	 * @see #composeFeatureFamily()
	 * @see #verifyClasses()
	 * @see #generateCode()
	 * @see SPLStructure#cleanup()
	 */
	public void compose(final boolean codegenEnabled)
			throws FileNotFoundException, IOException {

		if (!this.spl.getFeatureMap().isEmpty()) {

			this.renameClasses();

			this.composeFeatures();

			/* Benchmark: time needed to compose */
			long compEndTime = Timer.getInstance().end();
			System.out.println("composition: " + (compEndTime * 1e-6) + " ms");

			/* Benchmark: time needed to check for errors */
			Timer.getInstance().start();

			boolean errorsFound = this.verifyClasses();

			/* Benchmark: time needed to check for errors */
			long tcEndTime = Timer.getInstance().end();
			System.out.println("typecheck: " + (tcEndTime * 1e-6) + " ms");

			/* Benchmark: time needed to codegen */
			Timer.getInstance().start();

			// FIXME uncomment again!
			// if (codegenEnabled && !errorsFound) {

			this.generateCode();

			// }

			/* Benchmark: time needed to codegen */
			long codegenTime = Timer.getInstance().end();
			System.out.println("codegen: " + (codegenTime * 1e-6) + " ms");

		} else {

			logger.warn("The feature map is empty. Nothing to compose.");

		}

		this.spl.cleanup();

	}

	/**
	 * Renames the anonymous inner classes for each feature, so that the name
	 * conflict between two different anonymous classes within a common outer
	 * class is solved.
	 * 
	 * @throws IOException
	 *             if the class file could not be opened, altered or copied
	 * @throws FileNotFoundException
	 *             if the class file could not be found
	 */
	private void renameClasses() throws IOException, FileNotFoundException {

		logger.info("START RENAMING");

		for (Entry<String, List<String>> entry : this.spl.getFeatureMapInner()
				.entrySet()) {

			String featureName = entry.getKey();
			List<String> classFilePaths = entry.getValue();

			/* prepare replacements: [pattern] -> [replacement] */
			RenameUtil featRu = RenameUtil
					.generate(featureName, classFilePaths);

			/* DEBUG */
			logger.debug("Renaming classes of feature " + featureName + ": "
					+ featRu.toString() + ".");

			for (int i = 0; i < classFilePaths.size(); i++) {
				String sourceFilePath = classFilePaths.get(i);

				/* feature file corresponding to current class file */
				File featureFile = this.spl.getWorkingFile(featureName,
						sourceFilePath);

				logger.debug("Renaming class " + featureFile);

				/*
				 * replace all occurences with replacements within content of
				 * current class file
				 */

				/* read from file by first time ... */
				FileInputStream fis = new FileInputStream(featureFile);
				ClassReader cr = new ClassReader(fis);
				fis.close();

				/* make class file changeable */
				ClassWriter cw = new ClassWriter(cr, 0);

				/* prepare renaming tool chain */
				ClassRenameAdapter rca = new ClassRenameAdapter(cw, featRu);

				/* rename */
				cr.accept(rca, 0);

				/*
				 * if there was a replacement
				 */

				/* delete class file */
				featureFile.delete();

				/* fix file separators (file -> package) */
				String classPackagePath = FilePathUtils
						.separatorsToPackageStyle(sourceFilePath);
				/* rename class if needed */
				classPackagePath = featRu.renameClassFile(classPackagePath);

				/*
				 * if current file name does NOT change, it was a normal class
				 * file and it must be written to disk for further composition.
				 * ...
				 */
				if (FilePathUtils.separatorsToPackageStyle(sourceFilePath)
						.equals(classPackagePath)) {

					/* get processed file in working directory */
					File processedFile = this.spl.getWorkingFile(featureName,
							classPackagePath);
					// FileUtils.writeByteArrayToFile(processedFile,
					// cw.toByteArray());

					renamedFiles.put(processedFile, cw);

					// debug
					logger.debug("Finished renaming " + processedFile);
				}
				/*
				 * ... otherwise it is a nested anonymous file that can be moved
				 * to composition path (applies to inner classes only)
				 */
				else {
					/* create composition file with same name */
					File compFile = this.spl.getCompFile(classPackagePath);

					/* note finished class file */
					ClassNode node = ClassNodeTransformer
							.createClassNodeFrom(cw);
					this.finishedCompFiles.put(compFile, node);

					// debug
					logger.debug("Finished renaming " + compFile);
				} // end: if (sourceFilePath.equals(processedFilePath)) ... else

			} // end: for (int i = 0; i < classFilePaths.size(); i++)

		} // end: for (Entry<String, List<String>> entry :
			// this.spl.getFeatureMapAnonymous().entrySet())

		logger.info("END RENAMING");

	}

	/**
	 * Composes the feature files with the base files, following Fuji
	 * mechanisms. Also removes stubs from class files. Writes the result to the
	 * result folder.
	 * 
	 * @throws IOException
	 *             if a class file could not be opened, altered or copied
	 * @throws FileNotFoundException
	 *             if a class file could not be found
	 */
	private void composeFeatures() throws IOException, FileNotFoundException {

		logger.info("START COMPOSING USING " + this.getClass().getSimpleName());

		LinkedList<String> featureList = new LinkedList<String>(this.spl
				.getFeatureMap().keySet());
		// Collections.sort(l, Collections.reverseOrder());

		for (int i = 0; i < featureList.size(); i++) {

			String classFile = featureList.get(i);
			Queue<String> features = this.spl.getFeatureMap().get(classFile);

			/* debug */
			logger.debug("Composing " + classFile + " of features " + features);

			/*
			 * BASE FEATURE
			 */
			String baseName = features.poll();
			File baseFile = this.spl.getWorkingFile(baseName, classFile);
			ClassNode baseNode = getClassNodeFor(baseFile);
			logger.debug("loaded base class [stubs present: "
					+ this.stubsPresent + "]");

			// ignore enums
			if ((baseNode.access & Opcodes.ACC_ENUM) != 0) {
				features.clear();
			}

			/*
			 * FEATURES
			 */
			/*
			 * compose features
			 */
			while (!features.isEmpty()) {
				/* get feature file */
				String featureName = features.poll();
				File featureFile = this.spl.getWorkingFile(featureName,
						classFile);

				logger.debug("Refining " + classFile + " with feature "
						+ featureName);

				ClassNode featureNode = getClassNodeFor(featureFile);

				/*
				 * compose feature class with base class
				 */

				/* read feature class */
				ClassReader featureReader = ClassNodeTransformer
						.createClassReaderFrom(featureNode);
				/* make feature class changeable */
				ClassWriter featureWriter = new ClassWriter(featureReader,
						ClassWriter.COMPUTE_MAXS);
				/* use mergeadapter to merge base with feature class */
				ComposingClassVisitor featureAdapter = new ComposingClassVisitor(
						featureWriter, baseNode, baseName, this.stubsPresent);

				/* actually compose base with feature */
				featureReader.accept(featureAdapter, 0);

				/* get composed class */
				baseNode = ClassNodeTransformer
						.createClassNodeFrom(featureWriter);

				/* if result class was not created, throw exception */
				if (baseNode == null) {
					throw new IllegalStateException();
				}

				// do adjustments after each feature file
				baseNode = this.postCompositionSingleFeatureFile(baseNode);

				logger.debug("Composition successful");

				/*
				 * current feature is next base
				 */
				baseName = featureName;

			} // end: while (!features.isEmpty())

			/*
			 * END (all feature files composed)
			 */

			/* create file to merge into */
			File compFile = this.spl.getCompFile(classFile);

			/* note finished class file */
			this.finishedCompFiles.put(compFile, baseNode);

			/* info */
			logger.info("Composition of " + classFile + " finished");
		}

		/* info */
		logger.info("END COMPOSING USING " + this.getClass().getSimpleName());

	}

	/**
	 * TODO doc
	 * 
	 * @param classNode
	 * @return
	 */
	protected ClassNode postCompositionSingleFeatureFile(ClassNode classNode) {
		// do nothing
		return classNode;
	}

	/**
	 * TODO doc
	 * 
	 * @param baseFile
	 * @return
	 */
	protected ClassNode getClassNodeFor(File file) throws IOException {
		ClassNode baseNode;
		// remove stubs
		if (renamedFiles.containsKey(file)) {
			baseNode = StubUtil.removeAllStubs(renamedFiles.get(file));
		} else {
			baseNode = StubUtil.removeAllStubs(file);
		}

		return baseNode;
	}

	/**
	 * Verifies all the composed classes.
	 * 
	 * @return a boolean value whether verification was successful or not
	 */
	private boolean verifyClasses() {

		/* info */
		logger.info("START VERIFYING CLASSES with JustIce by Enver Haase");

		boolean errorFound = false;
		boolean warningFound = false;
		/*
		 * get all finished classes that need to be verified
		 */
		Set<ClassNode> finishedClassNodes = new HashSet<ClassNode>(
				this.finishedCompFiles.values());

		// FIXME remove sorting!
		ClassNode[] nodes = this.finishedCompFiles.values().toArray(
				new ClassNode[] {});
		Arrays.sort(nodes, new Comparator<ClassNode>() {
			@Override
			public int compare(ClassNode o1, ClassNode o2) {
				return o1.name.compareTo(o2.name);
			}
		});
		finishedClassNodes = new HashSet<ClassNode>(Arrays.asList(nodes));

		/*
		 * load all finished files into class path for verifying
		 */
		try {
			ClassLoader classPathClassLoader = ClassLoader
					.getSystemClassLoader();
			if (!classpath.isEmpty()) {
				File cpPath = new File(classpath);
				Collection<File> listFiles = FileUtils.listFiles(cpPath,
						new String[] { "jar", "class" }, true);

				classPathClassLoader = new JarFileClassLoader(listFiles);
			}
			/*
			 * start BCEL verifier
			 */
			BCELVerifier bcel = new BCELVerifier(finishedClassNodes,
					classPathClassLoader);

			bcel.verify();
			String trace = bcel.getTrace();
			errorFound = bcel.isErrorFound();
			warningFound = bcel.isWarningFound();

			// errors found or not
			if (errorFound) {
				logger.error("END VERIFYING CLASSES with errors.");
				logger.error(trace);
				errorFound = true;
			} else {
				logger.info("END VERIFYING CLASSES without errors.");
				if (warningFound) {
					logger.warn(trace);
				}
			}

		} catch (Exception e) {
			errorFound = true;
			logger.error(
					"An error occured during verification of the composed files!",
					e);
		}

		return errorFound;
	}

	/**
	 * TODO
	 * 
	 * @throws IOException
	 */
	private void generateCode() throws IOException {

		logger.info("START GENERATING CODE");

		for (Entry<File, ClassNode> entry : finishedCompFiles.entrySet()) {
			File classFile = entry.getKey();
			ClassNode classNode = entry.getValue();

			/* write into file */
			ClassWriter cw = ClassNodeTransformer
					.createClassWriterFrom(classNode);
			FileUtils.writeByteArrayToFile(classFile, cw.toByteArray());

			logger.debug("Generated file successfully: " + classFile);
		}

		logger.info("END GENERATING CODE");

	}

}
