package de.uni_passau.spl.featurebite;

import java.io.File;
import java.io.IOException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.log4j.Logger;

import de.uni_passau.spl.featurebite.spl.SPLStructure;
import de.uni_passau.spl.featurebite.util.FilePathUtils;
import de.uni_passau.spl.featurebite.util.Timer;

/**
 * The {@link Launcher} initializes the bytecode merging and its components.
 * 
 * @author Claus Hunsen
 */
public class Launcher {

	// TODO 70 - sequence diagram (e.g.,
	// http://asm.ow2.org/doc/developer-guide.html#classreader )

	/**
	 * The command line parser.
	 */
	private CommandLine cmd;

	/**
	 * The SPL's structure.
	 */
	private SPLStructure spl;

	/**
	 * The composer that works on the composing algorithm.
	 */
	private Composer composer;

	// TODO doc
	private final static Logger logger = Logger.getLogger(Launcher.class);

	/**
	 * Constructs a main objects that initializes the composing mechanism and
	 * puts together all necessary parts of it.
	 * 
	 * @param args
	 *            the program arguments
	 * 
	 * @throws IOException
	 * @throws ParseException
	 */
	public Launcher(String[] args) throws IOException, ParseException {

		/* Benchmark: measure time needed to create prerequisites */
		Timer.getInstance().start();

		/* Initialize options and parse the command line. */
		Options options = this.initOptions();
		this.cmd = new GnuParser().parse(options, args, true);

		/* help, version */
		if (this.cmd.hasOption(OptionNames.HELP)) {
			new HelpFormatter().printHelp(72, toolName(), "", options,
					logging(), true);
			return;
		} else if (this.cmd.hasOption(OptionNames.VERSION)) {
			System.out.println(version());
			return;
		}

		/* Check the features file argument. */
		if (this.cmd.getArgs().length != 1) {
			throw new MissingOptionException(
					"Exactly one features file must be specified. "
							+ this.cmd.getArgs() + "\n");
		}

		boolean dynamicEnabled = this.cmd.hasOption(OptionNames.DYNAMIC);

		/* Process SPL structure, parse sources and build ASTs. */
		File featuresFile = new File(this.cmd.getArgs()[0]);
		String compositionName = FilePathUtils.getFileName(featuresFile
				.getPath());

		File basedir;
		if (this.cmd.hasOption(OptionNames.BASEDIR)) {
			basedir = new File(this.cmd.getOptionValue(OptionNames.BASEDIR));
		} else {
			basedir = featuresFile.getCanonicalFile().getParentFile();
		}

		File destinationDir;
		if (this.cmd.hasOption(OptionNames.DESTDIR)) {
			destinationDir = new File(
					this.cmd.getOptionValue(OptionNames.DESTDIR));
		} else {
			destinationDir = new File(basedir, compositionName);
		}

		this.spl = new SPLStructure(featuresFile, compositionName, basedir,
				destinationDir, dynamicEnabled);

		String classpath = "";
		if (cmd.hasOption(OptionNames.CLASSPATH)) {
			classpath += cmd.getOptionValue(OptionNames.CLASSPATH);
		}

		/* Benchmark: time needed to create prerequisites */
		long taskCpuTimeNano = Timer.getInstance().end();
		System.out
				.println("prerequisites: " + (taskCpuTimeNano * 1e-6) + " ms");

		/* note start time */
		Timer.getInstance().start();

		if (!dynamicEnabled) {
			this.composer = new Composer(this.spl, classpath);
		} else {
			this.composer = new DynamicComposer(this.spl, classpath);
		}

		/* compose files */
		boolean codegenEnabled = !this.cmd.hasOption(OptionNames.NOCODEGEN);
		this.composer.compose(codegenEnabled);

	}

	private static String logging() {
		return "\nSet custom logger by loading your own log4j config file"
				+ " via \"-Dlog4j.configuration=<yourfile>\" on command line!\n"
				+ "Default level is WARN, logger prints to console.";
	}

	/**
	 * Initializes the options that can be triggered via command line.
	 * 
	 * @return the options
	 */
	@SuppressWarnings("static-access")
	private Options initOptions() {
		Options ops = new Options();
		ops.addOption(OptionBuilder.hasArg().withArgName("basedir")
				.withDescription("Specify where to find feature class files")
				.create(OptionNames.BASEDIR));
		ops.addOption(OptionBuilder
				.hasArg()
				.withArgName("directory")
				.withDescription("Specify where to place generated class files")
				.create(OptionNames.DESTDIR));
		ops.addOption(OptionBuilder
				.hasArg()
				.withArgName("path")
				.withDescription(
						"Specify where to find user class files and "
								+ "annotation processors (e.g., lib directory)")
				.withLongOpt("cp").create(OptionNames.CLASSPATH));
		ops.addOption(OptionBuilder
				.withDescription(
						"Composition mechanism removes stubs only if necessary and not all at once.")
				.create(OptionNames.DYNAMIC));
		ops.addOption(OptionBuilder.withDescription(
				"Do NOT generate .class files after composition process")
				.create(OptionNames.NOCODEGEN));
		ops.addOption(OptionBuilder.withDescription(
				"Print a synopsis of standard options")
				.create(OptionNames.HELP));
		ops.addOption(OptionBuilder.withDescription("Version information")
				.create(OptionNames.VERSION));
		return ops;
	}

	/**
	 * Enumerates all the options accepted by bcm.
	 */
	private static class OptionNames {
		public static final String HELP = "help"; //
		public static final String VERSION = "version";
		public static final String BASEDIR = "basedir"; //
		public static final String CLASSPATH = "classpath"; //
		public static final String DESTDIR = "d"; //
		public static final String DYNAMIC = "dynamic"; //
		public static final String NOCODEGEN = "nocodegen"; //
	}

	/**
	 * @return the tool's name
	 */
	private static String toolName() {
		return "FeatureBite";
	}

	// /**
	// * @return the tool's URL
	// */
	// private static String projectURL() {
	// // TODO ask for URL
	// return "http://www.example.org";
	// }

	/**
	 * @return the tool's version
	 */
	private static String version() {
		// TODO bump version
		return toolName() + " v.2013-06-12";
	}

	/**
	 * The main method.
	 * 
	 * @param args
	 *            the programm arguments
	 */
	public static void main(String[] args) {

		long appStartTime = Timer.getCpuTime();

		try {

			new Launcher(args);

		} catch (IOException exception) {
			logger.error("There was an error while reading the class files.",
					exception);
		} catch (ParseException exception) {
			logger.error("There was an error while parsing the command-line arguments.",
					exception);
		}

		long appEndTime = Timer.getCpuTime();
		System.out.println("complete: " + ((appEndTime - appStartTime) * 1e-6)
				+ " ms");

	}

}
