package de.uni_passau.spl.featurebite.adapter.composing;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.InnerClassNode;
import org.objectweb.asm.tree.MethodNode;

import de.uni_passau.spl.featurebite.adapter.detector.FieldAdapter;
import de.uni_passau.spl.featurebite.adapter.detector.MethodAdapter;
import de.uni_passau.spl.featurebite.adapter.rename.RenameUtil;
import de.uni_passau.spl.featurebite.adapter.rename.SuperClassRenameAdapter;
import de.uni_passau.spl.featurebite.util.NameMangling;

// * The {@link ComposingClassVisitor} merges a feature class with a given base
// * class - all mechanisms and rules as in <a
// * href="http://fosd.de/fuji">fuji</a>.<br />
// * The actually merged {@link ClassNode} is stored within the given ClassWriter
// * object. This ClassNode object still inherits all stub methods that were
// * generated in order to compile the feature class independently. Those (or
// * rather representing {@link FieldNode} objects) are stored and can be
// * retrieved via {@link #getStubFields()} and {@link #getStubFields()} and
// * removed via a {@link StubRemover}.<br />
// * The structure of merged class is as follows: feature fields and methods
// * (including stubs!) and then followed by base methods and fields. This allows
// * the {@link StubRemover} to work.
/**
 * TODO doc: stubs are removed first!
 * 
 * @author Claus Hunsen
 */
public class ComposingClassVisitor extends ClassVisitor {

	/**
	 * The name of an initializer method (constructor).
	 */
	public static final String INITIALIZER_NAME = "<init>";

	/**
	 * The name of a static initializer.
	 */
	public static final String STATIC_INITIALIZER_NAME = "<clinit>";

	/**
	 * The name of the underlying class. Both, base and feature class have the
	 * same name.
	 */
	private String className;

	/*
	 * BASE FEATURE
	 */

	/**
	 * The name of the base feature.
	 */
	private final String baseFeatureName;

	/**
	 * The ClassNode fully representing the base class.
	 */
	private ClassNode baseNode;

	/**
	 * The list of fields within the base class.
	 */
	private List<FieldNode> baseFields = new ArrayList<FieldNode>();

	/**
	 * The list of {@link FieldAdapter} instances for any feature field.
	 */
	private List<FieldAdapter> featFieldAdapters = new ArrayList<FieldAdapter>();

	/**
	 * The list of methods within the base class.
	 */
	private List<MethodNode> baseMethods = new ArrayList<MethodNode>();

	/**
	 * The list of constructors within the feature class.
	 */
	private List<MethodNode> baseStaticConstructors = new ArrayList<MethodNode>();

	/*
	 * FEATURE
	 */

	/**
	 * The list of inner classes within the feature class.
	 */
	private List<InnerClassNode> featInnerClasses = new ArrayList<InnerClassNode>();

	/**
	 * The list of constructors within the feature class.
	 */
	private List<ConstructorOriginalVisitor> featConstructorOriginalVisitors = new ArrayList<ConstructorOriginalVisitor>();

	// /**
	// * The list of {@link FieldAdapter} instances for any feature field.
	// */
	// private List<FieldAdapter> featFieldAdapters = new
	// ArrayList<FieldAdapter>();

	/**
	 * The list of {@link MethodAdapter} instances for any feature method.
	 */
	private List<MethodAdapter> featMethodAdapters = new ArrayList<MethodAdapter>();

	/*
	 * COMPOSITION
	 */

	/**
	 * Flag whether the name of the class was changed or not (based on
	 * composition rules).
	 */
	private boolean superNameChanged;

	/**
	 * The old super class name.
	 */
	private String superNameOld;

	/**
	 * The new super class name.
	 */
	private String superNameNew;

	private boolean stubsPresent;

	/*
	 * STUBS
	 */

	// /**
	// * The name of the refining feature.
	// */
	// private final String featureName;

	// /**
	// * The list of stub fields (or rather their mangled names) within the
	// * feature class.
	// */
	// private final List<String> stubFields = new ArrayList<String>();
	//
	// /**
	// * The list of stub methods (or rather their mangled names) within the
	// * feature class.
	// */
	// private final List<String> stubMethods = new ArrayList<String>();

	/**
	 * Creates a ComposingClassVisitor where the refining feature class is given
	 * via {@code featureWriter} and the base feature via {@code baseNode}.
	 * 
	 * @param featureWriter
	 *            ClassWriter instance that holds the feature class
	 * @param baseNode
	 *            the base class
	 * @param baseFeatureName
	 *            the name of the base feature
	 * @param stubsPresent
	 *            indicates whether stubs are present in base and feature class
	 */
	public ComposingClassVisitor(ClassWriter featureWriter, ClassNode baseNode,
			String baseFeatureName, boolean stubsPresent) {
		super(Opcodes.ASM4, featureWriter);
		this.baseNode = baseNode;
		this.baseFeatureName = baseFeatureName;
		// this.featureName = currentFeatureName;

		/* retrieve base fields */
		this.baseFields = this.baseNode.fields;
		/* retrieve base methods */
		this.baseMethods = this.baseNode.methods;

		/* are stubs present here? */
		this.stubsPresent = stubsPresent;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * Additionally, it modifies the the super class name accoording to the
	 * super class of the base class and the interface lists of feature and base
	 * are combined.
	 */
	@Override
	public void visit(int version, int access, String name, String signature,
			String superName, String[] interfaces) {

		/* saves the class's name */
		this.className = name;

		// TODO implement enum composition?
		// if ((access & Opcodes.ACC_ENUM) != 0) {
		// // int ACC_PUBLIC = 0x0001; // class, field, method
		// boolean isPublic = (access & Opcodes.ACC_PUBLIC) != 0;
		// // int ACC_PRIVATE = 0x0002; // class, field, method
		// // int ACC_PROTECTED = 0x0004; // class, field, method
		// // int ACC_STATIC = 0x0008; // field, method
		// boolean isStatic = (access & Opcodes.ACC_STATIC) != 0;
		// // int ACC_FINAL = 0x0010; // class, field, method
		// boolean isFinal = (access & Opcodes.ACC_FINAL) != 0;
		// // int ACC_SUPER = 0x0020; // class
		// boolean isSuper = (access & Opcodes.ACC_SUPER) != 0;
		// // int ACC_INTERFACE = 0x0200; // class
		// // int ACC_ABSTRACT = 0x0400; // class, method
		// boolean isAbstract = (access & Opcodes.ACC_ABSTRACT) != 0;
		// // int ACC_SYNTHETIC = 0x1000; // class, field, method
		// // int ACC_ANNOTATION = 0x2000; // class
		// // int ACC_ENUM = 0x4000; // class(?) field inner
		// boolean isEnum = (access & Opcodes.ACC_ENUM) != 0;
		// // int ACC_DEPRECATED = 0x20000; // class, field, method
		//
		// System.out.println(access + " " + this.className);
		// }

		/*
		 * modify access
		 * 
		 * class access is modified automatically since this classadapter visits
		 * the feature class and passes its access parameter to the super.visit
		 * method below
		 */

		/*
		 * modify superclass
		 * 
		 * if feature class' super class is Object and and base super class is
		 * some specific class (!=Object), the result super class is the base
		 * one. if feature super class is not Object, it is kept.
		 * 
		 * case1: Base extends AnyClass, Feature (refines Base) => Feature
		 * extends AnyClass;
		 * 
		 * case1b: Base extends Object, Feature (refines Base) => Feature
		 * extends Object;
		 * 
		 * case2: Base extends AnyClass, Feature extends SomeClass (refines
		 * Base) => Feature extends SomeClass;
		 * 
		 * TODO: log warning if SomeClass does not extend AnyClass (can cause
		 * java errors due to missing methods in the composed class)
		 */
		// XXX superName = "entenhausen/Donald";
		String objectType = Type.getInternalName(Object.class);
		if (superName.equals(objectType)
				&& !this.baseNode.superName.equals(objectType)) {
			this.superNameChanged = true;
			this.superNameOld = superName;

			superName = this.baseNode.superName;
			this.superNameNew = superName;
		}

		/*
		 * add interfaces
		 */
		// XXX this.baseNode.interfaces.add("entenhausen/Dagobert");
		Set<String> allInterfaces = new HashSet<String>(
				Arrays.asList(interfaces));
		allInterfaces.addAll(this.baseNode.interfaces);
		interfaces = Arrays.copyOf(allInterfaces.toArray(),
				allInterfaces.toArray().length, String[].class);

		/*
		 * visit this class
		 */
		super.visit(version, access, name, signature, superName, interfaces);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * Additionally, identifies if a field is a stub using a
	 * {@link FieldAdapter}. The visited feature field (whether stub or not) is
	 * accepted anyway.
	 * 
	 * @see #getBaseFieldForFeatureField(FieldNode)
	 * @see FieldAdapter
	 * @see ClassVisitor#visitField(int, String, String, String, Object)
	 */
	@Override
	public FieldVisitor visitField(int access, String name, String desc,
			String signature, Object value) {

		/* only needed for stub detection */
		FieldNode featFn = new FieldNode(access, name, desc, signature, value);

		/* get corresponding base method - if existing */
		FieldNode baseFn = this.getBaseFieldForFeatureField(featFn);

		/* obtain the field visitor so far */
		FieldVisitor fv = super
				.visitField(access, name, desc, signature, value);

		/*
		 * find out whether base and feature have to be merged or not
		 */
		if (baseFn != null) {

			fv = new FieldAdapter(fv, featFn);

			this.featFieldAdapters.add((FieldAdapter) fv);

		} // end: if (baseFn != null)

		/*
		 * accept current feature field node
		 */
		return fv;
	}

	/**
	 * Returns the corresponding base field (non-stub) from the list of base
	 * fields, {@code null} if there is none.
	 * 
	 * @param featFn
	 *            the feature field
	 * @return the corresponding base feature field
	 */
	private FieldNode getBaseFieldForFeatureField(FieldNode featFn) {

		FieldNode toReturn = null;

		/*
		 * find the corresponding base field to current feature field
		 */
		for (Iterator<FieldNode> it = this.baseFields.iterator(); it.hasNext();) {
			FieldNode baseFn = it.next();

			/*
			 * merge fields if base and feature field are the same; two fields
			 * are equal if name and type match.
			 */
			if (NameMangling.getFieldIdentifier(featFn.name, featFn.desc)
					.equals(NameMangling.getFieldIdentifier(baseFn.name,
							baseFn.desc))) {
				toReturn = baseFn;
				break;
			}
		}

		return toReturn;

	}

	/**
	 * {@inheritDoc}
	 * 
	 * Additionally, exceptions of base and refining method are merged together.
	 * If there was a super class change (see {@link #superNameChanged}), then
	 * the super call in the constructor is changed, too. (see
	 * {@link SuperClassRenameAdapter}) Merging of constructors (static and
	 * non-static) is treated via {@link ComposingStaticInitializerVisitor} and
	 * "normal" methods via a {@link MethodAdapter} (identifies if a method is a
	 * stub). The visited feature method is accepted anyway.
	 * 
	 * @see #getBaseMethodForFeatureMethod(MethodNode)
	 * @see ComposingStaticInitializerVisitor
	 * @see MethodAdapter
	 * @see SuperClassRenameAdapter
	 * @see ClassVisitor#visitMethod(int, String, String, String, String[])
	 */
	@Override
	public MethodVisitor visitMethod(int access, String name, String desc,
			String signature, String[] exceptions) {

		/*
		 * if method is a original stub method (name == original), skip it
		 */
		if (MethodAdapter.isOriginalMethod(name)) {
			return null;
		}

		/* only needed for original+stub detection */
		MethodNode featMn = new MethodNode(access, name, desc, signature,
				exceptions);

		/* get corresponding base method - if existing */
		MethodNode baseMn = this.getBaseMethodForFeatureMethod(featMn);

		/*
		 * EXCEPTIONS merge
		 */
		if (baseMn != null && baseMn.exceptions != null) {
			Set<String> allExceptions = new HashSet<String>();
			if (exceptions != null) {
				allExceptions.addAll(Arrays.asList(exceptions));
			}
			allExceptions.addAll(baseMn.exceptions);
			exceptions = Arrays.copyOf(allExceptions.toArray(),
					allExceptions.toArray().length, String[].class);
		}

		/*
		 * change ACCESS due to underlying base constructor
		 */
		if (baseMn != null && name.equals(INITIALIZER_NAME)) {
			access = baseMn.access;
			featMn.access = baseMn.access;
		}

		/* obtain the method visitor so far */
		MethodVisitor mv = this.cv.visitMethod(access, name, desc, signature,
				exceptions);

		/*
		 * SUPER CALL rename
		 * 
		 * rename super call in constructor if super class was changed in
		 * #visit(int, int, String, String, String, String[])
		 */
		if (this.superNameChanged) {
			/* identify that current method is constructor */
			if (name.equals(INITIALIZER_NAME)) {
				/*
				 * construct renaming util from "oldSuperClass" to
				 * "newSuperClass"
				 */
				RenameUtil ru = new RenameUtil(this.superNameOld,
						this.superNameNew);

				/* fit method adapter into visitor chain */
				mv = new SuperClassRenameAdapter(mv, ru);
			}
		}

		/*
		 * BODY
		 */
		if (baseMn != null) {

			/*
			 * if the current method is a INITIALIZER, then
			 * ConstructorOriginalVisitor handles the composition
			 */
			if (name.equals(INITIALIZER_NAME)) {

				// add call to original constructor
				mv = new ConstructorOriginalVisitor(mv, this.className,
						this.baseFeatureName, featMn, baseMn);

				this.featConstructorOriginalVisitors
						.add((ConstructorOriginalVisitor) mv);

			}

			/*
			 * if the current method is a STATIC INITIALIZER, then
			 * ComposingStaticInitializerVisitor handles the composition
			 */
			else if (name.equals(STATIC_INITIALIZER_NAME)) {
				// TODO 4 - align names of CSIV and COV!
				// FIXME use org.objectweb.asm.commons.StaticInitMerger?
				mv = new ComposingStaticInitializerVisitor(mv, featMn, baseMn);

				this.baseStaticConstructors.add(baseMn);
			}

			/*
			 * else a "NORMAL" method is refined or overwritten, so get a method
			 * adapter to determine whether to merge or not
			 */
			else {

				mv = new MethodAdapter(mv, featMn, this.baseFeatureName);

				/*
				 * add to global list, because it may be the case there was a
				 * original call. (see #visitEnd, part Methods for delegation
				 * and renaming actions)
				 */
				this.featMethodAdapters.add((MethodAdapter) mv);

			}
		} // end: if (baseMn != null)

		return mv;
	}

	/**
	 * Returns the corresponding base method (non-stub) from the list of base
	 * methods, {@code null} if there is none.
	 * 
	 * @param featMn
	 *            the feature method
	 * @return the corresponding base feature method
	 */
	private MethodNode getBaseMethodForFeatureMethod(MethodNode featMn) {

		MethodNode toReturn = null;

		/*
		 * find the corresponding base method to current feature method
		 */
		for (Iterator<MethodNode> it = this.baseMethods.iterator(); it
				.hasNext();) {
			MethodNode baseMn = it.next();

			/*
			 * merge methods if base and feature method are equal
			 */
			/*
			 * Normally: two methods are equal if return type and signature
			 * match. Here, the method node's signature already inherits the
			 * return type, so that only name and description have to match.
			 */
			String featMangled = NameMangling.getMethodIdentifier(featMn.name,
					featMn.desc);
			String baseMangled = NameMangling.getMethodIdentifier(baseMn.name,
					baseMn.desc);
			if (baseMangled.equals(featMangled)) {
				toReturn = baseMn;
				break;
			}
		}

		return toReturn;

	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see ClassVisitor#visitInnerClass(String, String, String, int)
	 */
	@Override
	public void visitInnerClass(String name, String outerName,
			String innerName, int access) {

		/*
		 * Inner classes have to be accepted since they here contain just a
		 * reference outer class that they are defined.
		 * 
		 * Because inner classes have their own .class files, they are merged by
		 * Merger class.
		 * 
		 * Inner classes that are contained in base class only (especially
		 * anonymous inner classes) are merged in #visitEnd().
		 */

		super.visitInnerClass(name, outerName, innerName, access);

		/*
		 * note that current inner class is defined
		 */
		this.featInnerClasses.add(new InnerClassNode(name, outerName,
				innerName, access));
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see #composeFields()
	 * @see #dumpAlreadyMergedConstructors()
	 * @see #composeMethods()
	 * @see #composeInnerClasses()
	 * @see ClassVisitor#visitEnd()
	 */
	@Override
	public void visitEnd() {

		/*
		 * FIELDS
		 * 
		 * add all base fields to merged class that are not composed with
		 * feature field
		 */

		this.composeFields();

		// /*
		// * CONSTRUCTORS
		// *
		// * filter constructors from base methods. they were merged above in
		// * #visitMethod.
		// */
		//
		// this.dumpAlreadyMergedConstructors();

		/*
		 * METHODS
		 * 
		 * add all base methods to merged class that are not composed with
		 * feature method
		 */

		this.composeMethods();

		/*
		 * INNER CLASSES
		 * 
		 * add all base inner classes to merged class
		 */
		this.composeInnerClasses();

		/*
		 * CLOSE CLASS
		 */

		super.visitEnd();

	}

	/**
	 * TODO
	 * 
	 * Identifies for each visited feature field whether it is a stub or not. If
	 * one is, it is marked as stub field in {@link #stubFields} and the
	 * corresponding (non-stub) base field (if exists) is accepted, too. If one
	 * feature field is not a stub, the corresponding base field is dumped. Any
	 * other base field left are accepted.
	 * 
	 * NOTE: The Strings for {@link #stubFields} must be generated using
	 * {@link NameMangling#getFieldIdentifier(String, String)}!
	 * 
	 */
	private void composeFields() {

		List<FieldNode> baseFieldsToRemove = new ArrayList<FieldNode>();

		for (Iterator<FieldNode> it = this.baseFields.iterator(); it.hasNext();) {
			FieldNode baseFn = it.next();
			String baseMangled = NameMangling.getFieldIdentifier(baseFn.name,
					baseFn.desc);

			for (Iterator<FieldAdapter> fvIt = this.featFieldAdapters
					.iterator(); fvIt.hasNext();) {
				FieldAdapter fa = fvIt.next();
				FieldNode featFn = fa.getFieldNode();

				/* find corresponding feature method if existing */
				String featMangled = NameMangling.getFieldIdentifier(
						featFn.name, featFn.desc);
				if (baseMangled.equals(featMangled)) {

					/*
					 * feat field is a stub
					 */
					if (this.stubsPresent && fa.isStubField()) {
						// do nothing, just do not remove base field here
						// call StubUtil.removeStubs() later!
					}

					/*
					 * remove base field if feat field is no stub
					 */
					else {
						baseFieldsToRemove.add(baseFn);
					}

				}

			}
		} // for (Iterator<FieldNode> it = this.baseFields.iterator();
			// it.hasNext();)

		/* remove fields from base that were identified above */
		this.baseFields.removeAll(baseFieldsToRemove);

		/* merge base methods with original call */
		for (Iterator<FieldNode> it = this.baseFields.iterator(); it.hasNext();) {
			FieldNode fn = it.next();
			fn.accept(this.cv);
		}
	}

	/**
	 * TODO
	 * 
	 * Identifies whether a feature method is a stub or not. If one is, it is
	 * marked as stub method in {@link #stubMethods} and the corresponding
	 * (non-stub) base method (if exists) is accepted. If it is not, it is
	 * checked if there was an original call identified before. If there was an
	 * original call, the corresponding base method is renamed and accepted.
	 * 
	 * Does NOT accept any base method that fulfills any of the following
	 * conditions:
	 * <ul>
	 * <li>It is a constructor that was merged before.</li>
	 * <li>Its corresponding refining feature method does not have a original
	 * call.</li>
	 * </ul>
	 * 
	 * NOTE: The Strings for {@link #stubMethods} must be generated using
	 * {@link NameMangling#getMethodIdentifier(String, String)}!
	 */
	private void composeMethods() {

		List<MethodNode> baseMethodsToRemove = new ArrayList<MethodNode>();

		for (Iterator<MethodNode> it = this.baseMethods.iterator(); it
				.hasNext();) {
			MethodNode baseMn = it.next();
			String baseMangled = NameMangling.getMethodIdentifier(baseMn.name,
					baseMn.desc);

			for (Iterator<MethodAdapter> mvIt = this.featMethodAdapters
					.iterator(); mvIt.hasNext();) {
				MethodAdapter ma = mvIt.next();
				MethodNode featMn = ma.getMethodNode();

				/* find corresponding feature method if existing */
				String featMangled = NameMangling.getMethodIdentifier(
						featMn.name, featMn.desc);
				if (baseMangled.equals(featMangled)) {

					/*
					 * static initializers were merged before in #visitMethod()
					 */
					if (baseMn.name.equals(STATIC_INITIALIZER_NAME)) {
						baseMethodsToRemove.add(baseMn);
						continue;
					}

					/*
					 * if feature method has original call rename base method
					 * and accept it
					 */
					if (ma.hasOriginalCall()) {
						baseMn.name = ma.getNewMethodName();
					}

					/*
					 * if feature method is a stub, note that and accept base
					 * method
					 */
					else if (this.stubsPresent && ma.isStubMethod()) {
						// do nothing, just do not remove base method here
						// call StubUtil.removeStubs() later!
					}

					/*
					 * methods without original call or stub pendant do not have
					 * to be merged
					 */
					else {
						baseMethodsToRemove.add(baseMn);
					}
				}
			}
		}

		/* remove methods without original call */
		this.baseMethods.removeAll(baseMethodsToRemove);

		// remove static initializers
		this.baseMethods.removeAll(this.baseStaticConstructors);

		/* merge base methods with original call */
		for (Iterator<MethodNode> it = this.baseMethods.iterator(); it
				.hasNext();) {
			MethodNode mn = it.next();
			mn.accept(this.cv);
		}
	}

	/**
	 * Accepts any base inner class that were not accepted before.
	 */
	private void composeInnerClasses() {

		/* those inner classes that are already accepted via feature class */
		List<InnerClassNode> baseInnerClassesToRemove = new ArrayList<InnerClassNode>();

		/* collect base inner classes that need to be accepted */
		for (Iterator<InnerClassNode> it = this.baseNode.innerClasses
				.iterator(); it.hasNext();) {
			InnerClassNode baseIcn = it.next();

			for (Iterator<InnerClassNode> featIt = this.featInnerClasses
					.iterator(); featIt.hasNext();) {
				InnerClassNode featIcn = featIt.next();

				/* find corresponding feature inner class if existing */
				if (baseIcn.name.equals(featIcn.name)) {
					baseInnerClassesToRemove.add(baseIcn);
				}
			}
		}
		/* remove methods without original call */
		this.baseNode.innerClasses.removeAll(baseInnerClassesToRemove);

		/* merge base inner classes that were not merged before */
		for (Iterator<InnerClassNode> it = this.baseNode.innerClasses
				.iterator(); it.hasNext();) {
			InnerClassNode icn = it.next();
			icn.accept(this.cv);
		}

	}
}
