package de.uni_passau.spl.featurebite.adapter.composing;

import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.commons.AdviceAdapter;
import org.objectweb.asm.commons.LocalVariablesSorter;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.MethodNode;

/**
 * The {@link ComposingStaticInitializerVisitor} merges a feature constructor
 * with a given base class constructor. It inserts the base constructor
 * statements plainly after the feature constructor's super call.
 * 
 * @author Claus Hunsen
 */
public class ComposingStaticInitializerVisitor extends AdviceAdapter {

	private final MethodNode baseConstructor;

	/**
	 * Creates a new adapter that chains itself into the visitor chain before
	 * the given {@link MethodVisitor} and inserts the statements of the given
	 * base constructor.
	 * 
	 * @param mv
	 *            the method visitor prior in visitor chain
	 * @param featConstructor
	 *            a {@link MethodNode} representing the visited feature
	 *            constructor
	 * @param baseConstructor
	 *            a {@link MethodNode} fully embodies the base constructor and
	 *            its statements
	 */
	public ComposingStaticInitializerVisitor(MethodVisitor mv, MethodNode featConstructor,
			MethodNode baseConstructor) {
		super(Opcodes.ASM4, mv, featConstructor.access, featConstructor.name, featConstructor.desc);

		/*
		 * strip base constructor
		 */
		this.baseConstructor = baseConstructor;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * Inserts the base constructor statements (except super call) into the
	 * feature constructor.
	 * 
	 * @see AdviceAdapter#onMethodEnter()
	 */
	@Override
	protected void onMethodEnter() {

		super.onMethodEnter();

		InsnList baseInsn = this.baseConstructor.instructions;

		/* reset labels */
		baseInsn.resetLabels();

		/*
		 * accept instructions into initializer (except return statements)
		 */
		for (int i = 0; i < baseInsn.size(); i++) {

			AbstractInsnNode node = baseInsn.get(i);

			// FIXME check other return-types: IRETURN, ...?
			if (node.getOpcode() == Opcodes.RETURN) {
				continue;
			}
			node.accept(this.mv);
		}

	}

	/**
	 * {@inheritDoc}
	 * 
	 * Increments the maximum values by the maximum values from base
	 * constructor.
	 * 
	 * @see LocalVariablesSorter#visitMaxs(int, int)
	 */
	@Override
	public void visitMaxs(int maxStack, int maxLocals) {
		super.visitMaxs(maxStack + this.baseConstructor.maxStack, maxLocals
				+ this.baseConstructor.maxLocals);
	}

	// /**
	// * Strips the given MethodNode from the super call.
	// *
	// * @param constructor
	// * the constructor to be stripped from super call
	// * @return a constructor method without super call
	// *
	// * @see MethodStripAdapter
	// */
	// private MethodNode stripConstructor(final MethodNode constructor) {
	//
	// /* construct new method node */
	// MethodNode strippedBaseConstructor = new MethodNode();
	//
	// /* strip constructor and store result in the newly constructed node */
	// MethodStripAdapter tca = new MethodStripAdapter(strippedBaseConstructor,
	// constructor);
	// constructor.accept(tca);
	//
	// /* reset the new constructor */
	// strippedBaseConstructor.instructions.resetLabels();
	//
	// /* return a stripped constructor */
	// return strippedBaseConstructor;
	// }
	//
	// /**
	// * A {@link MethodStripAdapter} strips a given {@link MethodNode},
	// * particularly a constructor, from its super call.
	// *
	// * @author Claus Hunsen
	// *
	// * @see <a
	// *
	// href="http://www.jroller.com/eu/entry/merging_class_methods_with_asm">http://www.jroller.com/eu/entry/merging_class_methods_with_asm</a>
	// *
	// */
	// public final class MethodStripAdapter extends AdviceAdapter {
	//
	// /**
	// * The method node that holds the stripped constructor.
	// */
	// private final MethodNode target;
	//
	// /**
	// * Constructs a new adapter that strips the given source constructor
	// * from its super call and stores the result in the target node.
	// *
	// * @param target
	// * a {@link MethodNode} that holds the result
	// * @param source
	// * a {@link MethodNode} that need to be stripped
	// */
	// public MethodStripAdapter(MethodNode target, MethodNode source) {
	// super(Opcodes.ASM4, target, source.access, source.name, source.desc);
	// this.target = target;
	// }
	//
	// /**
	// * Removes the super call.
	// */
	// @Override
	// protected void onMethodEnter() {
	// this.target.instructions.clear();
	// }
	//
	// /**
	// * Removes any return statements.
	// */
	// @Override
	// protected void onMethodExit(int opcode) {
	// // do nothing; skip return
	// }
	//
	// }

}
