package de.uni_passau.spl.featurebite.adapter.composing;

import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.commons.AdviceAdapter;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.MethodNode;

import de.uni_passau.spl.featurebite.adapter.rename.RenameUtil;

/**
 * @author Claus Hunsen
 * 
 */
public class ConstructorOriginalVisitor extends AdviceAdapter {

	/**
	 * TODO
	 */
	private final String className;

	/**
	 * TODO
	 */
	private final MethodNode baseConstructor;

	/**
	 * @param mv
	 *            the {@link MethodVisitor} containing the feature constructor
	 * @param className
	 *            the name of the class corresponding to the constructor
	 * @param baseFeatureName
	 *            the name of the feature prior to the current one
	 * @param featConstructor
	 *            the {@link MethodNode} representing the feature constructor
	 * @param baseConstructor
	 *            the {@link MethodNode} representing the base constructor
	 */
	public ConstructorOriginalVisitor(MethodVisitor mv, String className,
			String baseFeatureName, MethodNode featConstructor,
			MethodNode baseConstructor) {
		super(Opcodes.ASM4, mv, featConstructor.access, featConstructor.name,
				featConstructor.desc);
		this.className = className;
		this.baseConstructor = this.stripConstructor(baseConstructor);

		// rename base constructor
		this.baseConstructor.name = RenameUtil.renameClassNameByFeatureName(
				this.className, baseFeatureName);
		// transform to private method
		this.baseConstructor.access = ACC_PRIVATE;
	}

	/**
	 * @see org.objectweb.asm.commons.AdviceAdapter#onMethodEnter()
	 */
	@Override
	protected void onMethodEnter() {
		super.onMethodEnter();

		// load "this"
		this.mv.visitVarInsn(ALOAD, 0);

		this.fixConstructorArguments();

		// add call to original - now renamed - constructor
		this.mv.visitMethodInsn(Opcodes.INVOKESPECIAL, this.className,
				this.baseConstructor.name, this.baseConstructor.desc);
	}

	/**
	 * TODO doc
	 */
	private void fixConstructorArguments() {
		// load all arguments for calling the original constructor
		Type[] arguments = Type.getArgumentTypes(this.baseConstructor.desc);
		int count = 1;
		for (int i = 0; i < arguments.length; i++) {
			Type argument = arguments[i];
			String ad = argument.getDescriptor();

			// boolean || char || byte || short || int
			if (ad.equals(Type.BOOLEAN_TYPE.getDescriptor())
					|| ad.equals(Type.CHAR_TYPE.getDescriptor())
					|| ad.equals(Type.BYTE_TYPE.getDescriptor())
					|| ad.equals(Type.SHORT_TYPE.getDescriptor())
					|| ad.equals(Type.INT_TYPE.getDescriptor())) {
				this.mv.visitVarInsn(ILOAD, count);
			} else
			// long (64 bit = two stack slots!)
			if (ad.equals(Type.LONG_TYPE.getDescriptor())) {
				this.mv.visitVarInsn(LLOAD, count);
				count++;
			} else
			// float
			if (ad.equals(Type.FLOAT_TYPE.getDescriptor())) {
				this.mv.visitVarInsn(FLOAD, count);
			} else
			// double (64 bit = two stack slots!)
			if (ad.equals(Type.DOUBLE_TYPE.getDescriptor())) {
				this.mv.visitVarInsn(DLOAD, count);
				count++;
			} else
			// any object
			{
				this.mv.visitVarInsn(ALOAD, count);
			}

			count++;
		}

	}

	/**
	 * Strips the given MethodNode from the super call.
	 * 
	 * @param constructor
	 *            the constructor to be stripped from super call
	 * @return a constructor method without super call
	 * 
	 * @see MethodStripAdapter
	 */
	private MethodNode stripConstructor(final MethodNode constructor) {

		/**
		 * A {@link MethodStripAdapter} strips a given {@link MethodNode},
		 * particularly a constructor from its super call.
		 * 
		 * @author Claus Hunsen
		 */
		class MethodStripAdapter extends AdviceAdapter {

			/**
			 * The method node that holds the stripped constructor.
			 */
			private final MethodNode target;

			/**
			 * Constructs a new adapter that strips the given source constructor
			 * from its super call and stores the result in the target node.
			 * 
			 * @param target
			 *            a {@link MethodNode} that holds the result
			 * @param source
			 *            a {@link MethodNode} that need to be stripped
			 */
			public MethodStripAdapter(MethodNode target, MethodNode source) {
				super(Opcodes.ASM4, target, source.access, source.name,
						source.desc);
				this.target = target;
			}

			/**
			 * Removes the super call.
			 */
			@Override
			protected void onMethodEnter() {
				this.target.instructions.clear();
			}

			/**
			 * Removes any return statements.
			 */
			@Override
			protected void onMethodExit(int opcode) {
				// do nothing; skip return
			}

		}

		// construct new method node
		MethodNode strippedBaseConstructor = new MethodNode();

		// strip constructor and store result in the newly constructed node
		MethodStripAdapter tca = new MethodStripAdapter(
				strippedBaseConstructor, constructor);
		constructor.accept(tca);

		// reset the new constructor
		strippedBaseConstructor.instructions.resetLabels();

		// strip constructor
		InsnList newList = new InsnList();
		for (int i = 0; i < strippedBaseConstructor.instructions.size(); i++) {
			newList.add(strippedBaseConstructor.instructions.get(i));
		}
		constructor.instructions = newList;

		// return a stripped constructor
		return constructor;

	}

}
