package de.uni_passau.spl.featurebite.adapter.composing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.commons.AdviceAdapter;
import org.objectweb.asm.commons.LocalVariablesSorter;
import org.objectweb.asm.commons.Remapper;
import org.objectweb.asm.commons.RemappingMethodAdapter;
import org.objectweb.asm.commons.SimpleRemapper;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.TryCatchBlockNode;

/**
 * @author Claus Hunsen
 * 
 */
public class MethodCallInliner extends LocalVariablesSorter {
	private final String className;
	private final MethodNode baseConstructor;
	private List<TryCatchBlockNode> blocks = new ArrayList<TryCatchBlockNode>();
	private boolean inlining;
	private boolean inlined;

	MethodCallInliner(MethodVisitor mv, MethodNode featConstructor, MethodNode baseConstructor,
			String className) {
		super(featConstructor.access, featConstructor.desc, mv);
		this.className = className;
		this.baseConstructor = baseConstructor;
	}

	/**
	 * @see AdviceAdapter#onMethodEnter()
	 */

	@Override
	public void visitMethodInsn(int opcode, String owner, String name, String desc) {
		if (this.inlined) {
			this.mv.visitMethodInsn(opcode, owner, name, desc);
			return;
		}

		Map<String, String> map = Collections.singletonMap(this.className, this.className);
		Remapper remapper = new SimpleRemapper(map);
		Label end = new Label();
		this.inlining = true;
		this.baseConstructor.instructions.resetLabels();

		int access = opcode == Opcodes.INVOKESTATIC ? Opcodes.ACC_STATIC : 0;
		this.baseConstructor.accept(new InliningAdapter(this, end, access, desc, remapper));
		this.inlining = false;
		super.visitLabel(end);

		this.inlined = true;
	}

	@Override
	public void visitTryCatchBlock(Label start, Label end, Label handler, String type) {
		if (!this.inlining) {
			this.blocks.add(new TryCatchBlockNode(new LabelNode(start), new LabelNode(end),
					new LabelNode(handler), type));
		} else {
			super.visitTryCatchBlock(start, end, handler, type);
		}
	}

	@Override
	public void visitMaxs(int stack, int locals) {
		Iterator<TryCatchBlockNode> it = this.blocks.iterator();
		while (it.hasNext()) {
			TryCatchBlockNode b = it.next();
			super.visitTryCatchBlock(b.start.getLabel(), b.end.getLabel(), b.handler.getLabel(),
					b.type);
		}
		super.visitMaxs(stack, locals);
	}

	public static class InliningAdapter extends RemappingMethodAdapter {
		private final LocalVariablesSorter lvs;
		private final Label end;

		public InliningAdapter(LocalVariablesSorter mv, Label end, int access, String desc,
				Remapper remapper) {
			super(access, desc, mv, remapper);
			this.lvs = mv;
			this.end = end;
			int offset = (access & Opcodes.ACC_STATIC) != 0 ? 0 : 1;
			offset = 1;
			Type[] args = Type.getArgumentTypes(desc);
			for (int i = args.length - 1; i >= 0; i--) {
				super.visitVarInsn(args[i].getOpcode(Opcodes.ISTORE), i + offset);
			}
			if (offset > 0) {
				super.visitVarInsn(Opcodes.ASTORE, 0);
			}
		}

		@Override
		public void visitInsn(int opcode) {
			if (opcode == Opcodes.RETURN) {
				super.visitJumpInsn(Opcodes.GOTO, this.end);
			} else {
				super.visitInsn(opcode);
			}
		}

		@Override
		public void visitMaxs(int stack, int locals) {
		}

		@Override
		protected int newLocalMapping(Type type) {
			return this.lvs.newLocal(type);
		}
	}
}