package de.uni_passau.spl.featurebite.adapter.detector;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.MethodNode;

import de.uni_passau.spl.featurebite.adapter.rename.RenameUtil;
import de.uni_passau.spl.featurebite.stubs.StubUtil;

/**
 * The {@code MethodAdapter} is a common {@link MethodVisitor}, but it can
 * additionally identify if a method has a original call. Those calls are
 * automatically renamed using {@link #baseFeatureName} (see
 * {@link #getNewMethodName()} and {@link #ORIGINAL_RENAME_PATTERN} for further
 * information how the renaming is made and the new name is structured).
 * 
 * @author Claus Hunsen
 */
public class MethodAdapter extends MethodVisitor {

	/**
	 * The original method's name, usually "orginal".
	 */
	private static final String ORIGINAL_OPERATOR = "original";

	/**
	 * TODO like: original_$_(.*)
	 */
	private static final String NAMED_ORIGINAL_OPERATOR = String.format(
			RenameUtil.ORIGINAL_RENAME_PATTERN, ORIGINAL_OPERATOR, "");

	/**
	 * The representing {@link MethodNode} for the method examined using this
	 * {@link MethodAdapter} object.
	 * 
	 * NOTE: Only used for extracting {@link MethodNode#name}!
	 * 
	 * @see #visitMethodInsn(int, String, String, String)
	 */
	private final MethodNode mn;

	/**
	 * The name of underlying base feature (or previous feature in features
	 * list) used for renaming original calls.
	 * 
	 * @see #visitMethodInsn(int, String, String, String)
	 */
	private final String baseFeatureName;

	/**
	 * The renamed original call.
	 */
	private String newMethodCallName;

	/**
	 * Flag whether the examined method is a stub or not.
	 * 
	 * @see #visitAnnotation(String, boolean)
	 * @see #isStubMethod
	 * @see StubDetector
	 */
	private boolean isStubMethod;

	/**
	 * Flag whether the examined method has an original call or not.
	 * 
	 * @see #visitMethodInsn(int, String, String, String)
	 * @see #hasOriginalCall
	 */
	private boolean hasOriginalCall;

	/**
	 * TODO
	 * 
	 * @param methodName
	 * @return
	 */
	public static boolean isOriginalMethod(String methodName) {
		return (methodName.equals(ORIGINAL_OPERATOR) || methodName
				.startsWith(NAMED_ORIGINAL_OPERATOR));
	}

	/**
	 * Constructor that accepts the method visitor that is higher in visitor
	 * chain, a method node representing the method within the method visitor
	 * and a String that holds the base's feature name that is used for
	 * renaming.
	 * 
	 * @param mv
	 *            the {@link MethodVisitor} that is passed by ASM's visitor
	 *            calling chain
	 * @param mn
	 *            the {@link MethodNode} representive of the examined method
	 * @param baseFeatureName
	 *            the name of underlying base feature
	 */
	public MethodAdapter(MethodVisitor mv, MethodNode mn, String baseFeatureName) {
		super(Opcodes.ASM4, mv);
		this.mn = mn;
		this.baseFeatureName = baseFeatureName;

		this.isStubMethod = false;
		this.hasOriginalCall = false;
	}

	/**
	 * Visits the current annotation and checks if it is one that identifies
	 * stubs. Does not remove the annotation!
	 * 
	 * @see StubDetector#visitAnnotation(String, boolean)
	 * @see MethodVisitor#visitAnnotation(String, boolean)
	 */
	@Override
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {

		if (StubUtil.isStubAnnotation(desc)) {
			this.isStubMethod = true;
			/* remove annotation with: return null; */
		}

		return super.visitAnnotation(desc, visible);
	}

	/**
	 * Detects if a method instruction is a original call (
	 * {@link #ORIGINAL_OPERATOR}). If any, the call is renamed via with {link
	 * #ORIGINAL_RENAME_PATTERN} as pattern (new call name stored in
	 * {@link #newMethodCallName}) and {@link #hasOriginalCall()} set to
	 * {@code true}).
	 * 
	 * @see #hasOriginalCall
	 * @see #newMethodCallName
	 * @see #ORIGINAL_OPERATOR
	 * @see #ORIGINAL_RENAME_PATTERN
	 * 
	 * @see MethodAdapter#visitMethodInsn(int,String, String, String)
	 */
	@Override
	public void visitMethodInsn(int opcode, String owner, String name,
			String desc) {

		/*
		 * if original keyword is found, rename its call
		 */
		if (isOriginalMethod(name)) {
			// adapt return type (sometimes there is only one single
			// "Object giveIt()" method for all return types)
			desc = this.mn.desc;
			name = RenameUtil.renameMethodNameByFeatureName(this.mn.name,
					this.baseFeatureName);

			this.newMethodCallName = name;
			this.hasOriginalCall = true;
		}

		super.visitMethodInsn(opcode, owner, name, desc);
	}

	/**
	 * TODO doc
	 */
	public boolean isStubMethod() {
		return this.isStubMethod;
	}

	/**
	 * @return {@code true} if there was a original call; {@code false}
	 *         otherwise.
	 */
	public boolean hasOriginalCall() {
		return this.hasOriginalCall;
	}

	/**
	 * Returns the visited method node.
	 * 
	 * @return the current method as method node
	 */
	public MethodNode getMethodNode() {
		return this.mn;
	}

	/**
	 * Returns the new method name that previously was the original call.
	 * 
	 * @return the name of the base method that was renamed in
	 *         {@link #visitMethodInsn(int, String, String, String)}
	 */
	public String getNewMethodName() {
		return this.newMethodCallName;
	}

}
