package de.uni_passau.spl.featurebite.adapter.rename;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * The {@link ClassRenameAdapter} can make renaming of any String occurence
 * within the visited class (e.g., field types, return types and inner class
 * names). You provide a {@link RenameUtil} that holds the renaming pairs (old
 * -> new), and it substitutes accordingly as it goes.
 * 
 * @author Claus Hunsen
 */
public class ClassRenameAdapter extends ClassVisitor implements Opcodes {

	/**
	 * The utility that handles the renamings.
	 */
	private final RenameUtil ru;

	/**
	 * Creates a new adapter that decorates the given other visitor, and renames
	 * types using the given {@link RenameUtil}.
	 * 
	 * @param cv
	 *            the class visitor to decorate with renaming capability
	 * @param ru
	 *            the {@link RenameUtil} that handles any renaming action
	 */
	public ClassRenameAdapter(ClassVisitor cv, RenameUtil ru) {
		super(Opcodes.ASM4, cv);

		this.ru = ru;
	}

	/**
	 * Renames basic class identifier.
	 * 
	 * @see ClassVisitor#visit(int, int, String, String, String, String[])
	 */
	@Override
	public void visit(int version, int access, String name, String signature, String superName,
			String[] interfaces) {
		super.visit(version, access, this.ru.rename(name), this.ru.rename(signature),
				this.ru.rename(superName), this.ru.renameArray(interfaces));
	}

	/**
	 * Renames the type and signature of the visited field.
	 * 
	 * @see ClassVisitor#visitField(int, String, String, String, Object)
	 */
	@Override
	public FieldVisitor visitField(int access, String name, String desc, String signature,
			Object value) {

		desc = this.ru.renameType(desc);

		return super.visitField(access, name, desc, this.ru.rename(signature), value);
	}

	/**
	 * Renames return types and parameter types of the visited method. Also, it
	 * renames any occurences within the method.
	 * 
	 * @see MethodRenameAdapter
	 * @see ClassVisitor#visitMethod(int, String, String, String, String[])
	 */
	@Override
	public MethodVisitor visitMethod(int access, String name, String desc, String signature,
			String[] exceptions) {

		desc = this.ru.renameMethodType(desc);

		MethodVisitor mv = super.visitMethod(access, name, desc, this.ru.rename(signature),
				this.ru.renameArray(exceptions));

		if (mv != null) {
			mv = new MethodRenameAdapter(mv, this.ru);
		}

		return mv;
	}

	/**
	 * Renames inner classes within the current class.
	 * 
	 * @see ClassVisitor#visitInnerClass(String, String, String, int)
	 */
	@Override
	public void visitInnerClass(String name, String outerName, String innerName, int access) {
		super.visitInnerClass(this.ru.rename(name), this.ru.rename(outerName),
				this.ru.rename(innerName), access);
	}

	/**
	 * Renames any type of an existing annotation.
	 * 
	 * @see ClassVisitor#visitAnnotation(String, boolean)
	 */
	@Override
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
		desc = this.ru.renameType(desc);
		return super.visitAnnotation(this.ru.rename(desc), visible);
	}

	/**
	 * Renames any reference to an outer class.
	 * 
	 * @see ClassVisitor#visitOuterClass(String, String, String)
	 */
	@Override
	public void visitOuterClass(String owner, String name, String desc) {
		if (desc != null) {
			Type type = Type.getType(desc);
			switch (type.getSort()) {
			case Type.METHOD:
				desc = this.ru.renameMethodType(desc);
				break;

			case Type.OBJECT:
				// case Type.ARRAY:
				desc = this.ru.renameType(desc);
				break;

			default:
				desc = this.ru.rename(desc);
				break;
			}
		}

		super.visitOuterClass(this.ru.rename(owner), name, desc);
	}

	/**
	 * Renames the source of current class.
	 * 
	 * @see ClassVisitor#visitSource(String, String)
	 */
	@Override
	public void visitSource(String source, String debug) {
		super.visitSource(this.ru.renameClassFile(source), this.ru.rename(debug));
	}

}