package de.uni_passau.spl.featurebite.adapter.rename;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

/**
 * The {@link MethodRenameAdapter} can make renaming of any String occurence
 * within the visited method (e.g., types within type instructions, annotation
 * classes and types of local variables). You provide a {@link RenameUtil} that
 * holds the renaming pairs (old -> new), and it substitutes accordingly as it
 * goes.
 * 
 * @author Claus Hunsen
 */
class MethodRenameAdapter extends MethodVisitor {

	/**
	 * The utility that handles the renamings.
	 */
	private final RenameUtil ru;

	/**
	 * Creates a new adapter that decorates the given other visitor, and renames
	 * types using the given {@link RenameUtil}.
	 * 
	 * @param mv
	 *            the class visitor to decorate with renaming capability
	 * @param ru
	 *            the {@link RenameUtil} that handles any renaming action
	 */
	public MethodRenameAdapter(MethodVisitor mv, RenameUtil ru) {
		super(Opcodes.ASM4, mv);
		this.ru = ru;
	}

	/**
	 * Renames type within the stack map.
	 */
	@Override
	public void visitFrame(int type, int nLocal, Object[] local, int nStack, Object[] stack) {

		for (int i = 0; i < local.length; i++) {
			Object l = local[i];
			if (l instanceof String) {
				local[i] = this.ru.rename((String) l);
			}
		}
		for (int i = 0; i < stack.length; i++) {
			Object s = stack[i];
			if (s instanceof String) {
				stack[i] = this.ru.rename((String) s);
			}
		}

		super.visitFrame(type, nLocal, local, nStack, stack);
	}

	/**
	 * Renames types within type instructions.
	 */
	@Override
	public void visitTypeInsn(int i, String type) {
		this.mv.visitTypeInsn(i, this.ru.rename(type));
	}

	/**
	 * Renames types within field instructions.
	 */
	@Override
	public void visitFieldInsn(int opcode, String owner, String name, String desc) {
		desc = this.ru.renameType(desc);
		this.mv.visitFieldInsn(opcode, this.ru.rename(owner), name, desc);
	}

	/**
	 * Renames types within method instructions.
	 */
	@Override
	public void visitMethodInsn(int opcode, String owner, String name, String desc) {
		desc = this.ru.renameMethodType(desc);
		this.mv.visitMethodInsn(opcode, this.ru.rename(owner), name, desc);
	}

	/**
	 * Renames annotation types.
	 */
	@Override
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
		desc = this.ru.renameType(desc);
		return super.visitAnnotation(desc, visible);
	}

	/**
	 * Renames types of local variables.
	 */
	@Override
	public void visitLocalVariable(String name, String desc, String signature, Label start,
			Label end, int index) {
		desc = this.ru.renameType(desc);
		super.visitLocalVariable(name, desc, this.ru.rename(signature), start, end, index);
	}

	/**
	 * Renames types of new arrays.
	 */
	@Override
	public void visitMultiANewArrayInsn(String desc, int dims) {
		desc = this.ru.renameType(desc);
		super.visitMultiANewArrayInsn(desc, dims);
	}

	/**
	 * Renames parameter annotation types.
	 */
	@Override
	public AnnotationVisitor visitParameterAnnotation(int parameter, String desc, boolean visible) {
		desc = this.ru.renameType(desc);
		return super.visitParameterAnnotation(parameter, desc, visible);
	}

	/**
	 * Renames types in try-catch blocks.
	 */
	@Override
	public void visitTryCatchBlock(Label start, Label end, Label handler, String type) {
		super.visitTryCatchBlock(start, end, handler, this.ru.rename(type));
	}

}
