package de.uni_passau.spl.featurebite.adapter.rename;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.asm.Type;

import de.uni_passau.spl.featurebite.util.FilePathUtils;

/**
 * The {@link RenameUtil} is able to rename exact String occurences to new ones,
 * depending on a concrete renaming map.
 * 
 * There are three ways to instantiate a {@link RenameUtil}: use the two
 * constructors {@link #RenameUtil(String, String)} or {@link #RenameUtil(Map)},
 * or let it be constructed by the method {@link #generate(String, List)}. See
 * there for further instructions.
 * 
 * @author Claus Hunsen
 */
public final class RenameUtil {

	/**
	 * The pattern how an original call is renamed. Usually it is like
	 * {@code "_$_baseMethodName_baseFeatureName"}, where "baseMethodName" and
	 * "baseFeatureName" are parameters for e.g.
	 * {@link String#format(String, Object...)}.
	 * 
	 * @see #visitMethodInsn(int, String, String, String)
	 */
	public static final String ORIGINAL_RENAME_PATTERN = "%s_$_%s";
	
	/**
	 * The table that maps old to new names.<br />
	 * <br />
	 * NOTE: Both of these names are in "internal" form, meaning package names
	 * are separated by forward slash characters (/) instead of dots.
	 */
	public final Map<String, String> replacementTable;

	/**
	 * Constructs a {@link RenameUtil} that holds a single renaming pair.
	 */
	public RenameUtil(String oldName, String newName) {
		this.replacementTable = new HashMap<String, String>();
		this.replacementTable.put(oldName, newName);
	}

	/**
	 * Constructs a {@link RenameUtil} that holds the given map as renaming
	 * table.
	 * 
	 * @param renamingPairs
	 */
	public RenameUtil(Map<String, String> renamingPairs) {
		this.replacementTable = renamingPairs;
	}

	/**
	 * Replaces the given String with the correspondingone from the
	 * {@link #replacementTable}, if a mapping exists.
	 * 
	 * @param oldString
	 *            the String that is to be replaced
	 * @return the new String
	 */
	public String rename(String oldString) {

		if (oldString != null && this.replacementTable.containsKey(oldString)) {
			oldString = oldString.replace(oldString,
					this.replacementTable.get(oldString));
		}

		return oldString;

	}

	/**
	 * Replaces any String within the given array with the corresponding one
	 * from the {@link #replacementTable}, if a mapping exists.
	 * 
	 * @param oldStrings
	 *            the array whose Strings need to be renamed
	 * @return the array that holds the replaces Strings
	 */
	public String[] renameArray(String[] oldStrings) {

		if (oldStrings != null) {
			for (int i = 0; i < oldStrings.length; i++) {
				oldStrings[i] = this.rename(oldStrings[i]);
			}
		}

		return oldStrings;
	}

	/**
	 * Renames a class file path, ignoring any extension if one exists.
	 * 
	 * @param classFilePath
	 *            the class file path that need to be renamed
	 * @return the renamed path
	 */
	public String renameClassFile(String classFilePath) {

		if (classFilePath != null) {
			// /* save extension */
			// String extension = FilePathUtils.getExtension(classFilePath);

			/* save old file name */
			String oldClassName = FilePathUtils.removeExtension(classFilePath);

			/* rename the class regularly */
			String newClassName = this.rename(oldClassName);
			classFilePath = classFilePath.replace(oldClassName, newClassName);
		}

		return classFilePath;

	}

	// TODO doku
	public String renameType(String desc) {

		Type type = Type.getType(desc);

		switch (type.getSort()) {
		case Type.OBJECT:

			String renamedInternalName = this.rename(type.getInternalName());
			desc = Type.getObjectType(renamedInternalName).toString();

			break;

		case Type.ARRAY:
			Type elementType = type.getElementType();
			int dims = type.getDimensions();

			if (elementType.getSort() == Type.OBJECT) {
				String internalName = elementType.getInternalName();

				StringBuilder sb = new StringBuilder(this.rename(internalName));
				sb.insert(0, "L").append(";");
				for (int i = 0; i < dims; i++) {
					sb.insert(0, "[");
				}
				desc = sb.toString();
			}

			break;

		default:

			desc = this.rename(desc);

			break;
		}

		return desc;
	}

	// TODO doku
	public String renameMethodType(String desc) {

		Type returnType = Type.getReturnType(desc);
		Type[] argumentTypes = Type.getArgumentTypes(desc);

		// rename return type
		returnType = Type.getType(this.renameType(returnType.toString()));

		// rename parameter types
		for (int i = 0; i < argumentTypes.length; i++) {
			argumentTypes[i] = Type.getType(this.renameType(argumentTypes[i]
					.toString()));
		}

		desc = Type.getMethodDescriptor(returnType, argumentTypes).toString();

		return desc;
	}

	/**
	 * Returns a String representation of this object by appending the renaming
	 * table to the object's hash.
	 * 
	 * @see Object#toString()
	 */
	@Override
	public String toString() {
		return super.toString() + "=" + this.replacementTable.toString();
	}

	/**
	 * Generates a {@link RenameUtil} that can handle a renaming process for the
	 * given feature and the corresponding classes (or rather their files or
	 * paths).
	 * 
	 * @param featureName
	 *            the name of the feature whose classes are to be renamed
	 * @param classFiles
	 *            the classes that need to be renamed
	 * @return the {@link RenameUtil} that can handle the proper renaming
	 */
	public static RenameUtil generate(String featureName,
			List<String> classFiles) {

		// TODO clean-up

		/* replacements are [pattern] -> [replacement] */
		Map<String, String> replacementPairs = new HashMap<String, String>();

		for (int i = 0; i < classFiles.size(); i++) {
			/* tmp\foo\A$1.class */
			String classFilePath = classFiles.get(i);

			/*
			 * fix file separators (file -> package) + strip file extension =>
			 * tmp/foo/A$1
			 */
			String packageClassName = FilePathUtils
					.removeExtension(classFilePath);
			packageClassName = FilePathUtils
					.separatorsToPackageStyle(packageClassName);

			/* replacement is same as className (no change) - except if ... */
			String replacement = packageClassName;
			/* current class file is a inner class */
			if (replacement.contains("$")) {
				String[] oldParts = replacement.split("\\$");
				String[] parts = Arrays.copyOf(oldParts, oldParts.length);

				for (int j = 0; j < parts.length; j++) {
					String part = parts[j];

					if (part.matches("\\d+")) {
						/*
						 * inserts feature's name for uniqueness =>
						 * tmp/foo/A$featB1
						 */
						parts[j] = "" + featureName + "_" + parts[j];

					}

				}

				/*
				 * implode parts again and store intermediate results also
				 */
				StringBuilder oldSb = new StringBuilder();
				StringBuilder sb = new StringBuilder();

				oldSb.append(oldParts[0]);
				sb.append(parts[0]);
				replacementPairs.put(oldSb.toString(), sb.toString());

				for (int p = 1; p < parts.length; p++) {
					oldSb.append("$");
					oldSb.append(oldParts[p]);
					sb.append("$");
					sb.append(parts[p]);

					replacementPairs.put(oldSb.toString(), sb.toString());
					replacementPairs.put(oldParts[p], parts[p]);
				}

				// replacement = sb.toString();

			} // end: if (replacement.contains("$"))

		} // end: for (int i = 0; i < classFiles.size(); i++)

		RenameUtil ru = new RenameUtil(replacementPairs);
		return ru;
	}

	// TODO doku
	public static String renameClassNameByFeatureName(String className,
			String baseFeatureName) {
		String[] packagePath = className.split("\\/");
		String simpleClassName = String.format(ORIGINAL_RENAME_PATTERN,
				packagePath[packagePath.length - 1], baseFeatureName);
		return simpleClassName;
	}

	// TODO doku
	public static String renameMethodNameByFeatureName(String name,
			String baseFeatureName) {
		return String.format(ORIGINAL_RENAME_PATTERN, name, baseFeatureName);
	}

}
