package de.uni_passau.spl.featurebite.spl;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;

import de.uni_passau.spl.featurebite.util.FilePathUtils;

/**
 * A {@link SPLStructure} object the structure of an SPL, has full information
 * about features, roles and their relations.<br />
 * <br />
 * The (file and folder) structure before, while and after processing the
 * structure and composing the features is as follows:<br />
 * <br />
 * <strong>before</strong><br />
 * SamplePL &rarr; {@code base dir}<br />
 * |- base<br />
 * |- - A.java<br />
 * |- base_stubfix_classes &rarr; {@code feature source dir}<br />
 * |- - A.class &rarr; {@code feature source dir}<br />
 * |- feat<br />
 * |- - A.java<br />
 * |- feat_stubfix_classes<br />
 * |- - A.class<br />
 * |- - A$1.class<br />
 * |- SampleComp.features<br />
 * |- SampleComp_nonstubs.features<br />
 * <br />
 * The product line's name is <em>SamplePL</em>, the composition's name is
 * <em>SampleComp</em>.<br />
 * <br />
 * <strong>added while processing (deleted afterwards if error-free)</strong><br />
 * |- SampleComp_working &rarr; {@code working dir}<br />
 * |- |- base_stubfix_classes &rarr; {@code feature working dir}<br />
 * |- |- - A.class &rarr; {@code feature working file}<br />
 * |- |- feat_stubfix_classes<br />
 * |- |- - A.class<br />
 * |- |- - A$1.class<br />
 * <br />
 * <strong>added as result (will overwrite any existing folder of same
 * name)</strong><br />
 * |- SampleComp &rarr; {@code composition dir}<br />
 * |- - A.class &rarr; {@code composition file}<br />
 * |- - A$feat1.class &rarr; {@code composition file}<br />
 * |- - SimpleComp.features<br />
 * |- - SimpleComp.composition<br />
 * <br />
 * This {@code SPLStructure} object first copies the source files to a working
 * directory automatically, in order to ensure the integrity of the source
 * files. It also supports opening of any file or folder as needed (within this
 * working directory, but addressing is handled just with feature name and class
 * file name).<br />
 * <br />
 * The features and their files are held in so-called {@code feature maps}.
 * There is one for the main feature files and one for named and anonymous inner
 * classes. They are structered as follows:<br/>
 * <br />
 * <strong>featureMap (maps classes to contributing features)</strong><br />
 * <table style="width:200px;">
 * <tbody>
 * <tr>
 * <td style="vertical-align:top; font-weight: bold;">A.class &rarr;</td>
 * <td>
 * <ol>
 * <li>base</li>
 * <li>feat</li>
 * </ol>
 * </td>
 * </tr>
 * </tbody>
 * </table>
 * <br />
 * <br />
 * <strong>featureMapInner (maps features to containing inner classes + those
 * that may reference to them)</strong><br />
 * <table style="width:200px;">
 * <tr>
 * <td style="vertical-align:top; font-weight: bold;">feat &rarr;</td>
 * <td>
 * <ol>
 * <li>A$1.class</li>
 * <li>A.class</li>
 * </ol>
 * </td>
 * </tr>
 * </table>
 * <br />
 * The inner class feature map is needed because an inner classes of one feature
 * does not need to match an inner class of the same name within another
 * feature. This is due to the compilation process of Java itself, that assigns
 * consecutive numbers for the inner classes, always starting with 0.<br />
 * <br />
 * The object can be updated with the already composed files through the
 * {@link #addFinishedCompFile(File)} method.
 * 
 * @author Claus Hunsen
 */
public class SPLStructure {

	/**
	 * The name of the product line.
	 */
	private String name;

	/**
	 * The name of the called composition.
	 */
	private String compositionName;

	/**
	 * The base dir of this product line.
	 */
	private String basedirPath;

	/**
	 * The format of each feature folder.
	 */
	private static final String SOURCE_DIR_FORMAT = "{0}" + File.separator
			+ "{1}_stubfix_classes";

	/**
	 * The file extension of the feature classes.
	 */
	private static final String CLASS_FILE_EXT = "class";

	/**
	 * The pattern for a comment line within the .features file.
	 */
	private static final String FEATURE_FILE_COMMENT_PATTERN = "\\s*#.*";

	/**
	 * The format of the working dir.
	 */
	private static final String WORKING_DIR_FORMAT = "{0}_working"
			+ File.separator + "{1}";

	/**
	 * The pattern for a file of named inner class.
	 */
	public static final Pattern NAMEDINNERCLASS_PATTERN = Pattern.compile(
			".*\\$[^\\d]*." + CLASS_FILE_EXT, Pattern.CASE_INSENSITIVE);

	/**
	 * The format of each file within a feature working dir.
	 */
	private static final String WORKING_FILE_FORMAT = "{0}_working"
			+ File.separator + "{1}" + File.separator + "{2}";

	/**
	 * The list of all feature names that are also the names of the enclosing
	 * folder. (in the order they are listed in .features file)
	 */
	private List<String> featureNames;

	/**
	 * The .features file.
	 */
	private File featuresFile;

	/**
	 * The {@code featureMap}.
	 */
	private Map<String, Queue<String>> featureMap;

	/**
	 * The {@code featureMapInner}.
	 */
	private Map<String, List<String>> featureMapInner;

	/**
	 * The file name format for the nonstubs file.
	 */
	private static final String NONSTUBS_FILE_FORMAT = "{0}" + File.separator
			+ "{1}_nonstubs.txt";

	/**
	 * The split pattern for the lines within the nonstubs file.
	 */
	private static final String NONSTUBS_SPLIT_PATTERN = "(:)+";

	/**
	 * A table of all nonstubs mapping feature names to class files that are not
	 * stubs.
	 */
	private Map<String, List<String>> nonstubs;

	/**
	 * The format of each file within the composition directory.
	 */
	private static final String COMPOSITION_FILE_FORMAT = "{0}"
			+ File.separator + "{1}";

	/**
	 * The file extension of the .composition file.
	 */
	private static final String COMPOSITION_FILE_EXT = ".composition";

	// TODO doc
	private String destinationPath;

	// TODO doc
	private final static Logger logger = Logger.getLogger(SPLStructure.class);

	// TODO doc
	private boolean ignoreNonstubsFile;

	/**
	 * SPLStructure constructor setting up all needed directories and copies all
	 * needed files from source to working dir. Also initializes the feature
	 * maps and nonstubs.
	 * 
	 * @param featuresFile
	 *            absolute pathname of the features file.
	 * @param compositionName
	 * @param basedir
	 *            absolute pathname of the SPL's root directory
	 * @param destinationDir
	 *            TODO doc
	 * 
	 * @throws IOException
	 * @throws FeatureDirNotFoundException
	 */
	public SPLStructure(File featuresFile, String compositionName,
			File basedir, File destinationDir, boolean ignoreNonstubsFile)
			throws IOException, FeatureDirNotFoundException {

		/*
		 * direct assignments from parameters
		 */
		this.name = basedir.getName();
		this.basedirPath = basedir.getCanonicalPath();
		this.featuresFile = featuresFile;
		this.compositionName = compositionName;
		this.destinationPath = destinationDir.getCanonicalPath();
		this.ignoreNonstubsFile = ignoreNonstubsFile;

		/*
		 * directory preparations
		 */
		/* create (new) composition directory */
		this.getCompDirectory().mkdir();

		/*
		 * parse features file
		 */
		this.featureNames = this.parseFeaturesFile(this.featuresFile
				.getCanonicalPath());

		/* debug */
		logger.debug("FEATURES FILE: " + this.featuresFile.getAbsolutePath());
		logger.debug(this.featureNames);

		/*
		 * copy sources to working dir
		 */
		this.copySourcesToWorkingDir();

		/*
		 * initialize nonstubs and feature maps
		 */
		this.nonstubs = this.parseNonstubsFile();

		this.featureMap = new LinkedHashMap<String, Queue<String>>();
		this.featureMapInner = new LinkedHashMap<String, List<String>>();
		this.createFeatureMaps();

		/* debug */
		logger.debug("SPL STRUCTURE");
		logger.debug("nonstubs: " + this.nonstubs);
		logger.debug("featureMap: " + this.featureMap);
		logger.debug("featureMapAnonymous: " + this.featureMapInner);

	}

	/**
	 * Returns the {@code featureMap} for this product line composition.
	 * 
	 * @return the feature map
	 */
	public Map<String, Queue<String>> getFeatureMap() {
		return this.featureMap;
	}

	/**
	 * Returns the {@code featureMapInner} for this product line composition.
	 * 
	 * @return the feature map of inner and anonymous classes
	 */
	public Map<String, List<String>> getFeatureMapInner() {
		return this.featureMapInner;
	}

	/**
	 * Returns the source directory of the given feature as a {@link File}.
	 * 
	 * @param featureName
	 *            the name of the feature whose folder is to be retrieved
	 * 
	 * @return the folder as a {@code File}
	 * @throws FeatureDirNotFoundException
	 *             TODO doku
	 */
	private File getSourceDirectory(String featureName)
			throws FeatureDirNotFoundException {

		File sourceFile = new File(MessageFormat.format(SOURCE_DIR_FORMAT,
				this.basedirPath, featureName));

		return sourceFile;
	}

	/**
	 * Returns the working directory of the given feature as a {@link File}. If
	 * it does not exist, it will be created.
	 * 
	 * @param featureName
	 *            the name of the feature whose folder is to be retrieved
	 * 
	 * @return the folder as a {@code File}
	 */
	private File getWorkingDirectory(final String featureName) {

		File workingDir = new File(MessageFormat.format(WORKING_DIR_FORMAT,
				this.destinationPath, featureName));
		if (!workingDir.exists()) {
			workingDir.mkdir();
		}

		return workingDir;
	}

	/**
	 * Returns the working file of the given feature, that is describe through
	 * it class name, as a {@link File}.
	 * 
	 * @param featureName
	 *            the name of the feature whose folder is to be retrieved
	 * @param classFile
	 *            the name of the class file that is to be retrieved for the
	 *            given feature, given as a relative file path
	 * 
	 * @return the folder as a {@code File}
	 */
	public File getWorkingFile(String featureName, String classFile) {

		File workingFile = new File(MessageFormat.format(WORKING_FILE_FORMAT,
				this.destinationPath, featureName, classFile));

		if (!workingFile.exists()) {
			// sourceFile.mkdirs();
			// featureFile.createNewFile();
		}

		return workingFile;
	}

	/**
	 * Returns the composition directory of this product line composition as a
	 * {@link File}. If it does not exist, it will be created. If it exists, it
	 * will be deleted and re-created (overwritten).
	 * 
	 * @return the folder as a {@code File}
	 * 
	 * @throws IOException
	 *             in case the deletion could not be performed
	 */
	private File getCompDirectory() throws IOException {

		File compDir = new File(this.destinationPath);
		if (compDir.exists()) {
			FileUtils.forceDelete(compDir);
		}

		return compDir;

	}

	/**
	 * Returns the file located in the composition folder.
	 * 
	 * @param classFile
	 *            the path of the class file that is to be retrieved from the
	 *            composition folder, given as a relative file path
	 * 
	 * @return the class file as a {@code File} object
	 * 
	 * @throws IOException
	 *             in case the file could not be created
	 */
	public File getCompFile(String classFile) throws IOException {

		File mergeFile = new File(MessageFormat.format(COMPOSITION_FILE_FORMAT,
				this.destinationPath, classFile));

		return mergeFile;
	}

	/**
	 * Parses the .features file containing the configuration.
	 * 
	 * @param ffPathname
	 *            absolute pathname to the .features file.
	 * 
	 * @return list of strings representing canonical pathnames of project's
	 *         feature directories.
	 * @throws FeatureDirNotFoundException
	 */
	private List<String> parseFeaturesFile(String ffPathname)
			throws IOException, FeatureDirNotFoundException {

		// TODO 0 - inlinedoku
		ArrayList<String> featureModulePathnames = new ArrayList<String>();
		ArrayList<String> notDirs = new ArrayList<String>();
		Pattern commentPattern = Pattern.compile(FEATURE_FILE_COMMENT_PATTERN,
				Pattern.CASE_INSENSITIVE);
		BufferedReader in = new BufferedReader(new FileReader(ffPathname));
		String line;
		while ((line = in.readLine()) != null) {
			line = line.trim();
			if (commentPattern.matcher(line).matches() || line.isEmpty()) {
				continue;
			}

			/* Check if the specified feature dir exists. */
			File featureModuleDirectory = this.getSourceDirectory(line);
			// String featureModulePathname =
			// FilePathUtils.concatPaths(this.basedirPath, line + "_stubfix_classes");
			if (featureModuleDirectory.exists()
					&& featureModuleDirectory.isDirectory()) {
				featureModulePathnames.add(line);
			} else {
				notDirs.add(line);
			}
		}
		in.close();
		if (!notDirs.isEmpty()) {
			StringBuilder message = new StringBuilder("[" + ffPathname + "] "
					+ "There are no bytecode feature modules ("
					+ MessageFormat.format(SOURCE_DIR_FORMAT, this.name, "*")
					+ ") " + "for the following specified features:\n");
			for (String s : notDirs) {
				message.append("- " + s + "\n");
			}
			message.append("Those are ignored during composition.");
			logger.warn(message);
			// throw new FeatureDirNotFoundException(message);
		}
		if (featureModulePathnames.size() == 0) {
			throw new FeatureDirNotFoundException(
					"Feature file does not specify feature directories.");
		}
		return featureModulePathnames;
	}

	/**
	 * @return the nonstubs table mapping features to containing non-stub class
	 *         files
	 * 
	 * @throws IOException
	 *             in case the nonstubs file could not be opened and/or read
	 */
	private Map<String, List<String>> parseNonstubsFile() throws IOException,
			FeatureDirNotFoundException {

		// TODO 0 - inlinedoku

		if (this.ignoreNonstubsFile) {
			return null;
		}

		/* the file to be read */
		File nonstubsFile = new File(MessageFormat.format(NONSTUBS_FILE_FORMAT,
				this.basedirPath, this.name));

		/* the new map for the nonstubs */
		Map<String, List<String>> nonstubs = new LinkedHashMap<String, List<String>>();

		/* read lines from file */
		List<String> lines = FileUtils.readLines(nonstubsFile);

		/* do for each line a analysis */
		for (String line : lines) {
			line = line.trim();

			/* skip empty lines */
			if (line.isEmpty()) {
				continue;
			}

			/*
			 * extract base information from line
			 */
			/* split line into parts */
			String[] splitted = line.split(NONSTUBS_SPLIT_PATTERN);
			/* first part is feature's name */
			String featureName = splitted[0];

			/*
			 * implode file path's components for current nonstub and construct
			 * file path by doing this
			 */
			StringBuilder sb = new StringBuilder();

			splitted[1] = splitted[1].replace(".", File.separator);
			sb.append(splitted[1]);
			for (int i = 2; i < splitted.length; i++) {
				sb.append(File.separator);
				sb.append(splitted[i]);
			}
			sb.append("." + CLASS_FILE_EXT);

			String featureFile = sb.toString();

			/*
			 * list current nonstub
			 */
			List<String> rg = nonstubs.get(featureName);
			/*
			 * create a new list of nonstubs for this feature if not existing
			 * already
			 */
			if (rg == null) {
				rg = new LinkedList<String>();
			}
			/* add current file to feature's list */
			rg.add(featureFile);
			nonstubs.put(featureName, rg);

		} // end: for (String line : lines)

		return nonstubs;
	}

	/**
	 * Processes all the class files in the feature modules and structures them
	 * within the feature maps.
	 * 
	 * @throws IOException
	 *             in case the working dir could not be initialized properly
	 * 
	 * @see #featureMap
	 * @see #featureMapInner
	 */
	private void createFeatureMaps() throws IOException {

		// TODO 0 - inlinedoku

		for (String featureName : this.featureNames) {

			File featureWorkingDir = this.getWorkingDirectory(featureName);

			Collection<File> files = FileUtils.listFiles(featureWorkingDir,
					new String[] { CLASS_FILE_EXT }, true);

			/* Process all the files in the subdirs of the startdir. */
			for (Iterator<File> featureFileIt = files.iterator(); featureFileIt
					.hasNext();) {
				File currentFile = featureFileIt.next();

				// like "tmp/foo/A.class"
				String relativePath = featureWorkingDir.toURI()
						.relativize(currentFile.toURI()).getPath();
				/* fix file separators (file -> package) */
				relativePath = FilePathUtils
						.separatorsToSystemStyle(relativePath);

				/* if class is a nested one */
				if (relativePath.contains("$")) {
					this.handleInnerClass(featureName, relativePath);
				}

				/* skip stub files */
				if (!this.ignoreNonstubsFile
						&& this.isStubClass(featureWorkingDir, featureName,
								currentFile)) {
					continue;
				}

				/* add feature name to existing group */
				Queue<String> baseRG = this.featureMap.get(relativePath);
				if (baseRG == null) {
					baseRG = new LinkedList<String>();
					this.featureMap.put(relativePath, baseRG);
				}
				baseRG.add(featureName);

			} // end: for (Iterator<File> featureFileIt = files.iterator();
				// featureFileIt.hasNext();)

			/*
			 * remove duplicates from anonymous inner classes of this feature
			 */
			if (this.featureMapInner.containsKey(featureName)) {
				List<String> anonClasses = this.featureMapInner
						.get(featureName);
				Set<String> h = new HashSet<String>(anonClasses);
				List<String> newList = new LinkedList<String>(h);
				this.featureMapInner.put(featureName, newList);
			}

		} // end: for (String fmPathname : fmPathnames)

		/*
		 * clean up anonymous classes
		 */
		List<Entry<String, List<String>>> toRetain = new LinkedList<Entry<String, List<String>>>();
		for (Entry<String, List<String>> entry : this.featureMapInner
				.entrySet()) {

			List<String> classFiles = entry.getValue();

			/*
			 * if there is at least one unnamed inner class, renaming is
			 * necessary
			 */
			for (int i = 1; i < classFiles.size(); i++) {
				String classFile = classFiles.get(i);

				if (!NAMEDINNERCLASS_PATTERN.matcher(classFile).matches()) {
					toRetain.add(entry);
					break;
				}
			}

		}
		this.featureMapInner.entrySet().retainAll(toRetain);

	}

	/**
	 * Handles inner classes specially while processing the feature maps. Any
	 * inner class is added to the {@link #featureMapInner} to achieve this
	 * special treatment. If the current class file is a <em>named</em> inner
	 * class, it will be added to the nonstubs, too, in order to apply proper
	 * composition later.
	 * 
	 * @param featureName
	 *            the name of the current feature
	 * @param relativePath
	 *            the relative path to the current feature file
	 */
	private void handleInnerClass(String featureName, String relativePath) {

		/*
		 * compute the relative path of the outer class
		 */
		String outerClassRelativePath = relativePath
				.replaceAll("\\$.*\\.", ".");

		/*
		 * file represents a NAMED inner class that is NOT a stub
		 */
		if (NAMEDINNERCLASS_PATTERN.matcher(relativePath).matches()) {

			/* if outer class is a nonstub, inner class is also */
			if (this.nonstubs.containsKey(featureName)
					&& this.nonstubs.get(featureName).contains(
							outerClassRelativePath)) {
				this.nonstubs.get(featureName).add(relativePath);
			} else {
				List<String> rg = new LinkedList<String>();
				rg.add(relativePath);
				this.nonstubs.put(featureName, rg);
			}

		}

		/*
		 * file represents a inner class (anonymous or named)
		 */
		/*
		 * those anonymous classes does not care about the nonstubs because they
		 * are handled specially (just renaming); in contrary to named inner
		 * classes (see above).
		 */

		/* add feature name to existing group */
		List<String> baseRG = this.featureMapInner.get(featureName);
		if (baseRG == null) {
			baseRG = new LinkedList<String>();
		}

		/*
		 * add inner AND outer class to map (duplicates can be removed later)
		 */
		baseRG.add(outerClassRelativePath);
		baseRG.add(relativePath);
		this.featureMapInner.put(featureName, baseRG);

	}

	/**
	 * Determines whether a given class of a given feature is a stub file or not
	 * by looking up the file in the nonstubs table.
	 * 
	 * @param workingDir
	 *            the current working dir
	 * @param featureName
	 *            the current feature's name
	 * @param classFile
	 *            the currently processed class file within the given feature
	 * 
	 * @return <code>true</code> if the file is a stub, <code>false</code>
	 *         otherwise.
	 */
	private boolean isStubClass(File workingDir, String featureName,
			File classFile) {

		boolean isNotAStubClass = false;

		/* construct relative path like "tmp\foo\A.class" */
		String relativePath = workingDir.toURI().relativize(classFile.toURI())
				.getPath();
		/* fix file separators (package -> file) */
		relativePath = FilePathUtils.separatorsToSystemStyle(relativePath);

		/* if there is a nonstub entry is listed, it is a nonstub */
		if (this.nonstubs.containsKey(featureName)) {
			if (this.nonstubs.get(featureName).contains(relativePath)) {

				isNotAStubClass = true;

			}
		}

		return !isNotAStubClass;
	}

	/**
	 * Copies all needed (given by the .features file) feature source folders
	 * into the working directory.
	 * 
	 * @throws IOException
	 *             in case any file or directory could not be copied
	 */
	private void copySourcesToWorkingDir() throws IOException {

		/*
		 * copy features file
		 */
		File workingFeaturesFile = this.getWorkingFile("",
				this.featuresFile.getName());
		FileUtils.copyFile(this.featuresFile, workingFeaturesFile);

		/*
		 * copy bytecode feature folders
		 */
		for (String featureName : this.featureNames) {
				File sourceDir = this.getSourceDirectory(featureName);

				if (!sourceDir.exists()) {
					// logger.error("Could not find source directory: " + sourceFile);
					throw new FeatureDirNotFoundException(
							"The following feature directory does not exist in the SPL's feature directory:\n"
									+ sourceDir.getParentFile().getName()
									+ File.separator + sourceDir.getName());
				}
				
				File featureDir = this.getWorkingDirectory(featureName);
				featureDir.deleteOnExit();

				FileUtils.copyDirectory(sourceDir, featureDir);

		}

	}

	/**
	 * Writes a summary file into the composition folder containing the
	 * composition's structure data (feature maps and nonstubs).
	 * 
	 * @see #featureMap
	 * @see #featureMapInner
	 * @see #nonstubs
	 */
	public void writeCompositionFiles() {
		// TODO 0 - inlinedoku

		/*
		 * copy .features file
		 */
		try {

			File featuresFileComp = this.getCompFile(this.featuresFile
					.getName());

			// copy features file
			FileUtils.copyFile(this.featuresFile, featuresFileComp);

		} catch (IOException exception) {
			System.err.println("Features file " + this.featuresFile.getName()
					+ " could not be copied to composing folder!");
			exception.printStackTrace();
		}

		try {
			File compFile = this.getCompFile(this.compositionName
					+ COMPOSITION_FILE_EXT);

			String featureMapString = this.getFeatureMap().toString();
			String featureMapAnonymousString = this.getFeatureMapInner()
					.toString();

			// write into file
			List<String> lines = new ArrayList<String>();
			lines.add("feature map");
			lines.add(featureMapString);
			lines.add("");
			lines.add("feature map for inner classes");
			lines.add(featureMapAnonymousString);

			FileUtils.writeLines(compFile, lines, true);

		} catch (IOException exception) {
			System.err.println("Composition file " + this.compositionName
					+ COMPOSITION_FILE_EXT + " could not be written!");
			exception.printStackTrace();
		}

	}

	/**
	 * Cleans up the composition folders, deletes the working directory.
	 */
	public void cleanup() {

		try {
			File workingDir = this.getWorkingDirectory("");
			FileUtils.forceDelete(workingDir);
		} catch (IOException exception) {
			System.err.println("Deleting working directory did not work!");
			exception.printStackTrace();
		}

	}

}
