package de.uni_passau.spl.featurebite.stubs;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AnnotationNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.MethodNode;

import de.uni_passau.spl.featurebite.util.ClassNodeTransformer;
import de.uni_passau.spl.featurebite.util.CountMap;
import de.uni_passau.spl.featurebite.util.NameMangling;

/**
 * The feature .class files are filled up with stubs in order to make them
 * compilable. Those stubs (fields and methods) are annotated with the
 * annotation <code>@stub</code>. The {@link StubUtil} makes it possible to
 * interact with those stubs.
 * 
 * @author Claus Hunsen
 */
public class StubUtil {

	/**
	 * A list of all stub class internal names that were used.
	 */
	public static final List<String> STUB_ANNOTATION_TYPES = new ArrayList<String>();
	static {
		// current stub class
		STUB_ANNOTATION_TYPES.add(Type.getDescriptor(Stub.class));
		// old stub class 3
		STUB_ANNOTATION_TYPES
				.add("Lde/uni_passau/spl/bytecodecomposer/stubs/Stub;");
		// old stub class 2
		STUB_ANNOTATION_TYPES
				.add("Lde/uni_passau/spl/bytecodemerger/annotation/Stub;");
		// old stub class 1
		STUB_ANNOTATION_TYPES.add("Lclasslevel/annotation/Stub;");
		// tmp stub class
		STUB_ANNOTATION_TYPES.add("Ltmp/Stub;");
	}

	/**
	 * TODO doc
	 */
	public static boolean isStubAnnotation(String annotationDesc) {
		return STUB_ANNOTATION_TYPES.contains(annotationDesc);
	}
	
	/**
	 * Removes all stubs from the given .class file and returns the result as ClassNode.
	 * 
	 * @param classFile
	 *            the class file with stubs
	 * @return
	 * @throws IOException
	 *             if the class file could not be read or written
	 */
	public static ClassNode removeAllStubs(File classFile) throws IOException {
		/* create new class node and put base in it */
		ClassNode baseNode = ClassNodeTransformer
				.createClassNodeFrom(classFile);

		// remove stubs
		baseNode = removeAllStubs(baseNode);

		return baseNode;
	}
	
	/**
	 * Removes all stubs from the given  class writer instance and returns the result as ClassNode.
	 * 
	 * @param classWriter
	 *            the class writer with stubs
	 * @return
	 * @throws IOException
	 *             if the class file could not be read or written
	 */
	public static ClassNode removeAllStubs(ClassWriter classWriter) throws IOException {
		/* create new class node and put base in it */
		ClassNode baseNode = ClassNodeTransformer
				.createClassNodeFrom(classWriter);

		// remove stubs
		baseNode = removeAllStubs(baseNode);

		return baseNode;
	}

	/**
	 * Removes all stubs from the given class node and returns it.
	 * 
	 * @param classNode
	 *            the class node with stubs
	 * @return the stub-free class node
	 */
	public static ClassNode removeAllStubs(ClassNode classNode) {
		// TODO documentation + clean-up

		// remove annotated stub methods
		List<MethodNode> methods = new LinkedList<MethodNode>();
		for (Iterator<MethodNode> iterator = classNode.methods.iterator(); iterator
				.hasNext();) {
			MethodNode mn = iterator.next();

			if (mn.visibleAnnotations != null) {
				for (Iterator<AnnotationNode> iterator2 = mn.visibleAnnotations
						.iterator(); iterator2.hasNext();) {
					AnnotationNode an = iterator2.next();

					if (isStubAnnotation(an.desc)) {
						methods.add(mn);
						continue;
					}
				}
			}
		}
		classNode.methods.removeAll(methods);

		// remove annotated stub fields
		List<FieldNode> fields = new LinkedList<FieldNode>();
		for (Iterator<FieldNode> fnIterator = classNode.fields.iterator(); fnIterator
				.hasNext();) {
			FieldNode fn = fnIterator.next();

			if (fn.visibleAnnotations != null) {
				for (Iterator<AnnotationNode> anIterator = fn.visibleAnnotations
						.iterator(); anIterator.hasNext();) {
					AnnotationNode an = anIterator.next();

					if (isStubAnnotation(an.desc)) {
						fields.add(fn);
						continue;
					}
				}
			}
		}
		classNode.fields.removeAll(fields);

		return classNode;
	}

	/**
	 * Removes all unnecessary stubs.
	 * 
	 * @param classNode
	 *            the class node with stubs
	 * @param stubFields
	 *            mangled names of class fields
	 * @param stubMethods
	 *            mangled names of class methods
	 * @return the stub-"free" class node
	 * @throws IOException
	 */
	public static ClassNode removeStubsDynamically(ClassNode classNode) {
		// TODO documentation + clean-up

		// count methods and fields via name mangling
		CountMap counts = countMethodsAndFields(classNode);

		/*
		 * METHODS
		 */
		List<MethodNode> methods = new LinkedList<MethodNode>();

		// remove annotated stub methods
		for (Iterator<MethodNode> iterator = classNode.methods.iterator(); iterator
				.hasNext();) {
			MethodNode mn = iterator.next();
			String mnId = NameMangling.getMethodIdentifier(mn.name, mn.desc);

			if (counts.get(mnId).getValue() >= 2
					&& mn.visibleAnnotations != null) {
				for (Iterator<AnnotationNode> iterator2 = mn.visibleAnnotations
						.iterator(); iterator2.hasNext();) {
					AnnotationNode an = iterator2.next();

					if (isStubAnnotation(an.desc)) {
						methods.add(mn);
						counts.decrement(mnId);
						continue;
					}
				}
			} // if (stubMethods.contains(mnId) && mn.visibleAnnotations !=
				// null)
		}
		classNode.methods.removeAll(methods);
		methods.clear();
		
		/*
		 * FIELDS
		 */

		List<FieldNode> fields = new LinkedList<FieldNode>();
		// remove annotated stub fields
		for (Iterator<FieldNode> fnIterator = classNode.fields.iterator(); fnIterator
				.hasNext();) {
			FieldNode fn = fnIterator.next();
			String fnId = NameMangling.getFieldIdentifier(fn.name, fn.desc);

			if (counts.get(fnId).getValue() >= 2
					&& fn.visibleAnnotations != null) {
				for (Iterator<AnnotationNode> anIterator = fn.visibleAnnotations
						.iterator(); anIterator.hasNext();) {
					AnnotationNode an = anIterator.next();

					if (isStubAnnotation(an.desc)) {
						fields.add(fn);
						counts.decrement(fnId);
						continue;
					}
				}
			} // if (counts.get(fnId).getValue() >= 2 && fn.visibleAnnotations != null)
		}
		classNode.fields.removeAll(fields);
		fields.clear();

		/*
		 * RETURN
		 */
		
		return classNode;
	}

	public static CountMap countMethodsAndFields(
			ClassNode classNode) {

		CountMap count = new CountMap();

		// count methods
		for (Iterator<MethodNode> iterator = classNode.methods.iterator(); iterator
				.hasNext();) {
			MethodNode mn = iterator.next();
			String mnId = NameMangling.getMethodIdentifier(mn.name, mn.desc);
			count.increment(mnId);
		}

		// count fields
		for (Iterator<FieldNode> fnIterator = classNode.fields.iterator(); fnIterator
				.hasNext();) {
			FieldNode fn = fnIterator.next();
			String fnId = NameMangling.getFieldIdentifier(fn.name, fn.desc);
			count.increment(fnId);
		}

		return count;
	}

}
