package de.uni_passau.spl.featurebite.verification;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;

import org.apache.bcel.Repository;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.util.ByteSequence;
import org.apache.bcel.util.ClassLoaderRepository;
import org.apache.bcel.verifier.VerificationResult;
import org.apache.bcel.verifier.Verifier;
import org.apache.bcel.verifier.VerifierFactory;
import org.apache.log4j.Logger;
import org.objectweb.asm.tree.ClassNode;

import de.uni_passau.spl.featurebite.util.ClassNodeTransformer;

public class BCELVerifier {

	// TODO doc
	private final static Logger logger = Logger.getLogger(BCELVerifier.class);

	private Collection<ClassNode> classNodes;

	boolean verified;

	StringBuilder trace;

	private boolean errorFound;
	private boolean warningFound;

	public BCELVerifier(Collection<ClassNode> classNodes, ClassLoader loader)
			throws ClassFormatException, IOException {
		this.classNodes = classNodes;
		this.verified = false;
		this.errorFound = false;
		this.warningFound = false;
		this.trace = new StringBuilder();

		ClassLoaderRepository clr = new ClassLoaderRepository(loader);
		Repository.setRepository(clr);
		loadClassesToRepository(classNodes, clr);
	}

	public BCELVerifier(ClassNode classNode, ClassLoader loader)
			throws ClassFormatException, IOException {
		this(Arrays.asList(classNode), loader);
	}

	public String getTrace() {
		if (verified) {
			return trace.toString();
		}

		throw new IllegalStateException(
				"Method #verify() must be called first!");
	}

	public boolean isErrorFound() {
		return errorFound;
	}

	public boolean isWarningFound() {
		return warningFound;
	}

	/**
	 * System.out.println("JustIce by Enver Haase, (C) 2001-2002.\n" +
	 * "<http://bcel.sourceforge.net>\n" + "<http://jakarta.apache.org/bcel>");
	 */
	public void verify() {

		/*
		 * verify each class
		 */
		for (ClassNode cn : this.classNodes) {
			String className = cn.name;
			className = className.replace('/', '.');
boolean errorFoundLocally = false;
			
			try {
				logger.info("Now verifying: " + className);
				trace.append("\n" + "-----------------\n" + "Now verifying: "
						+ className + "\n");

				Verifier verifier = VerifierFactory.getVerifier(className);
				VerificationResult vr;

				vr = pass1(verifier);
				vr = pass2(verifier);

				if (vr.getStatus() != VerificationResult.VERIFIED_OK) {
					errorFoundLocally = true;
					continue;
				}

				JavaClass jc = Repository.lookupClass(className);
				for (int methodIndex = 0; methodIndex < jc.getMethods().length; methodIndex++) {
					vr = pass3a(verifier, jc, methodIndex);
					vr = pass3b(verifier, jc, methodIndex);

					if (vr.getStatus() != VerificationResult.VERIFIED_OK) {
						errorFoundLocally = true;
					}
				}

				/* warnings */
				String[] warnings = verifier.getMessages();
				if (warnings.length != 0) {
					trace.append("Warnings for " + jc.getClassName() + ":\n");
					for (String warn : warnings) {
						trace.append(warn + "\n");
					}

					this.warningFound = true;
				}

				// avoid swapping.
				verifier.flush();
				// DO NOT DO THIS: Repository.clearCache(); !!!
				// System.gc();

				/* if cache is not empty, print errors */
				if (errorFoundLocally) {
					trace.append("Verification failed for " + cn.name + "!");
				}
				/* otherwise verification is successful */
				else {
					trace.append("Verification successful for " + cn.name + "!");
					errorFound = true;
				}

				// } catch (ClassNotFoundException ex) {
				// errorTrace.append("Verification failed for " + cn.name
				// + "\n" + ex.getStackTrace() + "\n");
			} catch (Exception ex) {
				trace.append("Verification failed for " + cn.name + "!" + "\n"
						+ ex.getStackTrace() + "\n");
				errorFound = true;
			}
		}// for (ClassNode cn : this.classNodes)

		verified = true;
	}

	private VerificationResult pass1(Verifier v) {
		VerificationResult vr = v.doPass1();

		if (!vr.equals(VerificationResult.VR_OK)) {
			trace.append("Pass 1:\n" + vr);
		}

		return vr;
	}

	private VerificationResult pass2(Verifier v) {
		VerificationResult vr = v.doPass2();

		if (!vr.equals(VerificationResult.VR_OK)) {
			trace.append("Pass 2:\n" + vr);
		}

		return vr;
	}

	private VerificationResult pass3a(Verifier v, JavaClass jc, int methodIndex) {
		VerificationResult vr = v.doPass3a(methodIndex);

		if (!vr.equals(VerificationResult.VR_OK)) {
			trace.append("Pass 3a, method number " + methodIndex + " ['"
					+ jc.getMethods()[methodIndex] + "']:\n" + vr);
		}

		return vr;
	}

	private VerificationResult pass3b(Verifier v, JavaClass jc, int methodIndex) {

		VerificationResult vr = v.doPass3b(methodIndex);

		if (!vr.equals(VerificationResult.VR_OK)) {
			trace.append("Pass 3b, method number " + methodIndex + " ['"
					+ jc.getMethods()[methodIndex] + "']:\n" + vr);
		}

		return vr;
	}

	private void loadClassesToRepository(Collection<ClassNode> classNodes,
			ClassLoaderRepository clr) throws ClassFormatException, IOException {
		for (ClassNode cn : classNodes) {

			byte[] code = ClassNodeTransformer.createClassWriterFrom(cn)
					.toByteArray();

			ByteSequence bs = new ByteSequence(code);
			ClassParser parser = new ClassParser(bs, cn.name.replace('/', '.'));
			JavaClass jc = parser.parse();

			Repository.addClass(jc);
			logger.debug("add class \"" + jc.getClassName()
					+ "\" to verifier repository.");
		}
	}

}
